/*
 * $Id: viconfig.c,v 1.3 1998/07/29 23:46:47 pw Exp $
 * Copyright (C) 1998 Chris Csanady and Pete Wyckoff
 *
 * Make the ioctl which connect/disconnect a specific ether device to the
 * vi interface.
 */
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include "vivar.h"

static void usage(void);

int
main(int argc, char *argv[])
{
    char ch, *dev_name;
    int fd, error, cflag, uflag, iflag;
    extern char *optarg;
    extern int optind;

    cflag = uflag = iflag = 0;
    while ((ch = getopt(argc, argv, "hc:u:i:")) != -1)
	switch (ch) {
	case 'c':
	    cflag = 1;
	    dev_name = optarg;
	    break;
	case 'u':
	    uflag = 1;
	    dev_name = optarg;
	    break;
	case 'i':
	    iflag = 1;
	    dev_name = optarg;
	    break;
	case 'h':
	default:
	    usage();
	    return 1;
	}
    argc -= optind;
    argv += optind;

    if (cflag + uflag + iflag != 1) {
	usage();
	return 1;
    }

    if ((fd = open(dev_name, O_RDWR, 0)) < 0) {
	printf("Could not open device %s\n", dev_name);
	return fd;
    }

    if (cflag) {
	const char *if_name = *argv;
	if (argc != 1) usage();
	printf("Attaching %s to %s\n", if_name, dev_name);
	if ((error = ioctl(fd, VI_IFATTACH, if_name)) < 0) {
	    perror("Attaching");
	    return error;
	}
    }
    if (uflag) {
	if (argc != 0) usage();
	printf("Detaching from %s\n", dev_name);
	if ((error = ioctl(fd, VI_IFDETACH, NULL)) < 0) {
	    perror("Detaching");
	    return error;
	}
    }
    if (iflag) {
	struct vi_info_req req;

	if (argc != 0) usage();
	if ((error = ioctl(fd, VI_INFO, &req)) < 0) {
	    perror("Getting info");
	    return error;
	}
#ifdef __linux__
	printf("%s: attached to %s\n", dev_name, req.if_name);
#else
	printf("%s: attached to %s%d\n", dev_name, req.if_name, req.if_unit);
#endif
    }
    (void) close(fd);
    return 0;
}

static void
usage()
{
    fprintf(stderr, "Usage: viconfig -c <device> <interface> : attach\n");
    fprintf(stderr, "       viconfig -u <device>             : detach\n");
    fprintf(stderr, "       viconfig -i <device>             : get info\n");
}

