/*
 * fxp-output
 * Copyright (C) 1998 Chris Csanady and Pete Wyckoff
 */
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/socket.h>
#include <sys/mbuf.h>
#include <vm/vm.h>              /* for vtophys */
#include <vm/pmap.h>            /* for vtophys */
#include <net/if.h>
#if __FreeBSD__ > 2
#include <net/if_var.h>         /* for ifunit() */
#include <net/if_arp.h>
#else
#include <netinet/in.h>
#include <netinet/if_ether.h>
#endif
#include <net/if_media.h>
#include <net/ethernet.h>
#include <pci/if_fxpreg.h>
#include <pci/if_fxpvar.h>
#include "vivar.h"

#define FXP_NTXCB       128

int tx_threshold = 64;

int driver_output(struct vi_cseg *, struct vi_instance *);

/*
 * Start packet transmission on the interface.
 */
int
driver_output(struct vi_cseg *cs, struct vi_instance *vi)
{
    struct vi_dseg *ds = (struct vi_dseg *)(cs + 1);
    struct fxp_softc *sc = vi->ifp->if_softc;
    struct fxp_cb_tx *txp;
    struct vi_hdr *vh;
    struct mbuf *m;
    caddr_t data, next;
    int i, len, done = 1;

    if (sc->tx_queued >= FXP_NTXCB)
	return 1;

    txp = sc->cbl_last->next;

    MGET(m, M_TEMP, M_NOWAIT);
    if (m == NULL)
	return 1;
    bcopy(&vi->eh_template, m->m_data, sizeof(struct ether_header));
    vh = (struct vi_hdr *)(m->m_data + sizeof(struct ether_header));

    txp->tbd[0].tb_addr = vtophys(m->m_data);
    txp->tbd[0].tb_size = sizeof(struct ether_header) + sizeof(struct vi_hdr);

    if (ds->Length > vi->mtu) {
	len = min(ds->Length - vi->send_off, vi->mtu);
	data = vi_utok(vi, (caddr_t)ds->Data.Address + vi->send_off,
			ds->Handle, len);

	vh->svi = vi->svi;
	vh->dvi = vi->dvi;
	vh->off = vi->send_off;

	vi->send_off += len;
	if (vi->send_off < ds->Length)
	    done = 0;
    } else {
	vh->svi = vi->svi;
	vh->dvi = vi->dvi;
	
	len = ds->Length;
	data = vi_utok(vi, ds->Data.Address, ds->Handle, len);
    }
    vh->len = len;
    vh->flags = 0;

    if (data == NULL)
	return 0;

    txp->tbd[1].tb_addr = vtophys(data);
    next = (caddr_t)((vm_offset_t)data + PAGE_SIZE & ~PAGE_MASK);
    if (data + len > next) {
	txp->tbd[1].tb_size = next - data;
	txp->tbd[2].tb_size = len - txp->tbd[1].tb_size;
	txp->tbd[2].tb_addr = vtophys(next);
	txp->tbd_number = 3;
    } else {
	txp->tbd[1].tb_size = len;
	txp->tbd_number = 2;
    }

    txp->mb_head = m;
    txp->cb_status = 0;
    txp->cb_command =
	FXP_CB_COMMAND_XMIT | FXP_CB_COMMAND_SF | FXP_CB_COMMAND_S;
    txp->tx_threshold = tx_threshold;

    /*
     * Advance the end of list forward.
     */
    sc->cbl_last->cb_command &= ~FXP_CB_COMMAND_S;
    sc->cbl_last = txp;

    /*
     * Advance the beginning of the list forward if there are
     * no other packets queued (when nothing is queued, cbl_first
     * sits on the last TxCB that was sent out).
     */
    if (sc->tx_queued == 0)
	sc->cbl_first = txp;

    sc->tx_queued++;

    i = 10000;
    while (CSR_READ_1(sc, FXP_CSR_SCB_COMMAND) && --i);
    CSR_WRITE_1(sc, FXP_CSR_SCB_COMMAND, FXP_SCB_COMMAND_CU_RESUME);

    if (done) {
	vi->send_off = 0;
	return 0;
    }

    return 1;
}

