/*
 * mbuf-output
 * Copyright (C) 1998 Chris Csanady and Pete Wyckoff
 */
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/socket.h>
#include <sys/mbuf.h>
#include <net/if.h>
#if __FreeBSD__ > 2
#include <net/if_var.h>         /* for ifunit() */
#include <net/if_arp.h>
#else
#include <netinet/in.h>
#include <netinet/if_ether.h>
#endif
#include <net/ethernet.h>
#include "vivar.h"

int driver_output(struct vi_cseg *, struct vi_instance *);

/*
 * Start packet transmission on the interface.
 */
int
driver_output(struct vi_cseg *cs, struct vi_instance *vi)
{
    struct vi_dseg *ds = (struct vi_dseg *)(cs + 1);
    struct vi_hdr *vh;
    struct mbuf *mb_head, *m;
    caddr_t data, next_page;
    int i, len, s, done = 1;

again:
    MGETHDR(mb_head, M_DONTWAIT, MT_DATA);
    if (mb_head == NULL)
	goto failed;
    bcopy(&vi->eh_template, mb_head->m_data, sizeof(struct ether_header));
    vh = (struct vi_hdr *)(mb_head->m_data + sizeof(struct ether_header));
    mb_head->m_len = sizeof(struct ether_header) + sizeof(struct vi_hdr);

    MGET(mb_head->m_next, M_DONTWAIT, MT_DATA);
    if (mb_head->m_next == NULL)
	goto failed;
    m = mb_head->m_next;

    if (ds->Length > vi->mtu) {
	len = min(ds->Length - vi->send_off, vi->mtu);
	m->m_data = vi_utok(vi, (caddr_t)ds->Data.Address + vi->send_off,
			    ds->Handle, len);

	vh->svi = vi->svi;		/* These 3 fields should take	*/
	vh->dvi = vi->dvi;		/* one instruction using 64 bit	*/
	vh->off = vi->send_off;		/* loads.			*/

	vi->send_off += len;
	if (vi->send_off < ds->Length)
	    done = 0;
    } else {
	vh->svi = vi->svi;
	vh->dvi = vi->dvi;
	vh->off = 0;

	len = ds->Length;
	m->m_data = vi_utok(vi, ds->Data.Address, ds->Handle, len);
    }
    vh->len = len;	/* Header setup is now complete. */
    vh->flags = 0;

    if (m->m_data == NULL) {
	printf("mbuf_output: data not within memory handle.\n");
	m_freem(mb_head);
	return -1;
    }

    next_page = (caddr_t)((vm_offset_t)m->m_data + PAGE_SIZE & ~PAGE_MASK);
    if (m->m_data + len > next_page) {
	MGET(m->m_next, M_DONTWAIT, MT_DATA);
	if (m->m_next == NULL)
	    goto failed;
	m->m_len = next_page - m->m_data;
	m->m_next->m_len = len - m->m_len;
	m->m_next->m_data = next_page;
    } else {
	m->m_len = len;
    }

    mb_head->m_pkthdr.len = len + sizeof(struct ether_header) +
			    sizeof(struct vi_hdr);
	    
    s = splimp();
    IF_ENQUEUE(&vi->ifp->if_snd, mb_head);
    (*vi->ifp->if_start)(vi->ifp);
    splx(s);

    if (done) {
	vi->send_off = 0;
	return 0;
    }

    return 1;

failed:
    printf("mbuf_output: failed.\n");
    if (mb_head)
	m_freem(mb_head);
    return 1;
}
