/*
 * yfg_output
 * Copyright (C) 1998 Chris Csanady and Pete Wyckoff
 */
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/socket.h>
#include <sys/mbuf.h>
#include <vm/vm.h>              /* for vtophys */
#include <vm/pmap.h>            /* for vtophys */
#include <net/if.h>
#if __FreeBSD__ > 2
#include <net/if_var.h>         /* for ifunit() */
#include <net/if_arp.h>
#else
#include <netinet/in.h>
#include <netinet/if_ether.h>
#endif
#include <net/if_media.h>
#include <net/ethernet.h>
#include <pci/if_yfgreg.h>
#include "vivar.h"

#define YFG_NTXCB       128

int driver_output(struct vi_cseg *, struct vi_instance *);

/*
 * Start packet transmission on the interface.
 */
int
driver_output(struct vi_cseg *cs, struct vi_instance *vi)
{
    struct vi_dseg *ds = (struct vi_dseg *)(cs + 1);
    struct yfg_softc *sc = vi->ifp->if_softc;
    struct yfg_tx_cb *txp;
    struct vi_hdr *vh;
    struct mbuf *m;
    caddr_t data, next;
    int len, done = 1;

    if (sc->tx_queued >= YFG_NTXCB)
	return 1;

    txp = sc->cbl_last->next;

    MGET(m, M_TEMP, M_NOWAIT);
    if (m == NULL)
	return 1;
    bcopy(&vi->eh_template, m->m_data, sizeof(struct ether_header));
    vh = (struct vi_hdr *)(m->m_data + sizeof(struct ether_header));

    txp->tdb[0].addr = vtophys(m->m_data);
    txp->tdb[0].reqCount = sizeof(struct ether_header) + sizeof(struct vi_hdr);
    txp->tdb[0].cmd = CMD_OUT_MORE;

    if (ds->Length > vi->mtu) {
	len = min(ds->Length - vi->send_off, vi->mtu);
	data = vi_utok(vi, (caddr_t)ds->Data.Address + vi->send_off,
			ds->Handle, len);

	vh->svi = vi->svi;
	vh->dvi = vi->dvi;
	vh->off = vi->send_off;

	vi->send_off += len;
	if (vi->send_off < ds->Length)
	    done = 0;
    } else {
	vh->svi = vi->svi;
	vh->dvi = vi->dvi;
	
	len = ds->Length;
	data = vi_utok(vi, ds->Data.Address, ds->Handle, len);
    }
    vh->len = len;
    vh->flags = 0;

    if (data == NULL) {
	printf("yfg_output: data not within memory handle.\n");
	m_free(m);
	return -1;
    }

    txp->tdb[1].addr = vtophys(data);
    txp->tdb[1].cmd = CMD_OUT_MORE;
    next = (caddr_t)((vm_offset_t)data + PAGE_SIZE & ~PAGE_MASK);
    if (data + len > next) {
	txp->tdb[1].reqCount = next - data;
	txp->tdb[2].reqCount = len - txp->tdb[1].reqCount;
	txp->tdb[2].addr = vtophys(next);
	txp->tdb[2].cmd = CMD_OUT_LAST | BRAN_ALWAYS;
    } else {
	txp->tdb[1].reqCount = len;
	txp->tdb[1].cmd = CMD_OUT_LAST | BRAN_ALWAYS;
    }

    txp->mb_head = m;

    txp->tdb[YFG_NTXSEG - 1].cmd = CMD_STOP;

    /*
     * Advance the end of list forward.
     */
    sc->cbl_last->tdb[YFG_NTXSEG - 1].baddr = vtophys(&txp->tdb[0]);
    sc->cbl_last->tdb[YFG_NTXSEG - 1].cmd = CMD_NOP | BRAN_ALWAYS;
    sc->cbl_last = txp;

    /*
     * Advance the beginning of the list forward if there are
     * no other packets queued (when nothing is queued, cbl_first
     * sits on the last TxCB that was sent out).
     */
    if (sc->tx_queued == 0)
	sc->cbl_first = txp;

    sc->tx_queued++;

    sc->csr->TxControl = SET_MASK(WAKE, WAKE);

    if (done) {
	vi->send_off = 0;
	return 0;
    }

    return 1;
}

