/*
 * VIA-like psuedo-device driver
 * Copyright (C) 1998 Chris Csanady and Pete Wyckoff
 *
 * $Id: vi.c,v 1.26 1998/09/22 22:43:01 pw Exp $
 */

/*
 * Note:  (linux) Appropriate places to patch for via_input:
 *     drivers/net/eepro100.c
 * 	- stick in driver itself, my temp solution
 *     netif_rx
 * 	- gets called from drivers with skb
 * 	- drops the packet if backlog count is too big
 * 	- else just queues it on the tail of backlog, and marks NET_BH
 *     net_bh
 * 	- called if marked above, maybe the delay to here is Chris's problem
 * 	- can hack it in here
 *     dev_add_pack
 * 	- more general way to add a new protocol
 * 	  static struct packet_type vi_packet_type = {
 * 	    0,
 * 	    NULL,  -- any device, else a "tap" protocol
 * 	    vi_input,
 * 	    NULL,  -- data : ptr to the packet type
 * 	    NULL,  -- next field
 * 	  };
 * 	  vi_packet_type.type = htons(ETH_P_VI);
 * 	  dev_add_pack(&vi_packet_type);
 * 	int vi_input(struct sk_buff *skb, struct device *dev,
 * 	             struct packet_type *ptype);
 */
#ifdef __FreeBSD__
#include "vi.h"  /* NVI definition */
#else
#define NVI 1    /* fake it for linux */
#endif

#if (defined(__FreeBSD__) && NVI > 0) || defined(__linux__)

#ifdef __FreeBSD__
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/conf.h>
#include <sys/errno.h>
#include <sys/mbuf.h>
#include <sys/malloc.h>
#include <sys/socket.h>
#include <sys/queue.h>
#ifdef DEVFS
#include <sys/devfsext.h>
#endif
#include <vm/vm.h>
#include <vm/vm_prot.h>
#include <vm/pmap.h>
#include <vm/vm_extern.h>
#include <vm/vm_page.h>
#include <machine/segments.h>
#include <machine/md_var.h>
#include <net/if.h>
#if __FreeBSD__ > 2
#include <net/if_var.h>		/* for ifunit() */
#include <net/if_arp.h>
#else
#include <netinet/in.h>
#include <netinet/if_ether.h>
#endif
#include <net/ethernet.h>
#endif  /* FreeBSD includes */

#ifdef __linux__
#include <linux/config.h>
#ifdef MODULE
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/netdevice.h>
#include <linux/malloc.h>
#include <linux/vmalloc.h>
#include <asm/uaccess.h>
#endif  /* linux includes */

#include "vivar.h"

#define VI_DEBUG 0

#ifdef __FreeBSD__
static d_open_t		vi_open;
static d_close_t	vi_close;
static d_ioctl_t	vi_ioctl;

static struct cdevsw vi_cdevsw = {
    vi_open,		vi_close,	noread,		nowrite,
    vi_ioctl,		nullstop,	nullreset,	nodevtotty,
    seltrue,		nommap,		nostrat,	"vi",
    NULL,		-1
};
#ifndef DEVFS
static dev_t vi_cdevhandle;
#endif

/* Trap related stuff */
#define IDTVEC(name) __CONCAT(X,name)
extern void	IDTVEC(vi_trap);
extern void	IDTVEC(rsvd);

void		vi_service(struct vi_trap_args *);
__inline void	vi_send(struct vi_instance *, struct vi_cseg *);
__inline void	vi_recv(struct vi_instance *, struct vi_cseg *);
void		vi_input(struct mbuf *);
void		vi_free(caddr_t, u_int);
static int	vi_mapmh(struct vi_memhandle *);
static void	vi_unmapmh(struct vi_memhandle *);

/* Non-private for the benefit of libkvm. */
struct vi_softc		*vi_softc;
struct vi_stats		vi_stats;

/* The VI's are shared between all of the interfaces. */
struct vi_instance	**vis;	
int			num_vi;		/* number of open VI's */

extern vm_map_t		kernel_map;
extern void		(*via_input)(struct mbuf *);

#endif  /* FreeBSD header stuff */

#ifdef __linux__
static int vi_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
  unsigned long arg);
static int vi_open(struct inode *inode, struct file *file);
static int vi_close(struct inode *inode, struct file *file);

static struct file_operations vi_fops = {
    NULL,  /* llseek */
    NULL,  /* read */
    NULL,  /* write */
    NULL,  /* readdir */
    NULL,  /* poll */
    vi_ioctl,
    NULL,  /* mmap */
    vi_open,
#define VERSION(v,p,s) (((v)<<16)+(p<<8)+s)
#if (LINUX_VERSION_CODE >= VERSION(2,1,118))
    NULL,  /* flush */
#endif
    vi_close,
    NULL,  /* fsync */
    NULL,  /* fasync */
    NULL,  /* check_media_change */
    NULL,  /* revalidate */
    NULL,  /* lock */
};

asmlinkage void vi_trap(void);

#define MIN(a,b) ((a)<(b)?(a):(b))

/* bsd compatibility items */
#define bzero(d,l) memset(d,0,l)
#define free(a,v) kfree(a)

#if defined(TRAP_SYSTEM_CALL)
void		vi_service(struct vi_trap_args *);
#else
void		vi_service(void);  /* other ways to get the arg */
#endif
static void vi_input(struct sk_buff *skb);
static void	vi_send(struct vi_instance *, struct vi_cseg *);
static void	vi_recv(struct vi_instance *, struct vi_cseg *);
void		vi_free(caddr_t, u_long);
static int	vi_mapmh(struct vi_memhandle *);
static void	vi_unmapmh(struct vi_memhandle *);

/* Non-private for the benefit of libkvm. */
struct vi_softc		*vi_softc;
struct vi_stats		vi_stats;

/* The VI's are shared between all of the interfaces. */
struct vi_instance	**vis;	
int			num_vi;		/* number of open VI's */

#define printf printk
#define malloc(a,b,c) kmalloc(a,GFP_KERNEL)

#ifdef __alpha__
#  define LXORX "%lx"
#else
#  define LXORX "%x"
#endif

#endif  /* linux header stuff */

/*
 * Shared header stuff
 */
/* Called by main() during pseudo-device attachment, or module init */
static void vi_attach(void *);
#define DEFAULT_CDEV_MAJOR 100
static int cdev_major = DEFAULT_CDEV_MAJOR;

#define NUM_HARDWARE (10)
static struct hardware_desc *hardware[NUM_HARDWARE];

int
register_hardware(hw_type type, struct hardware_desc *desc)
{
    if (type < 0 || type >= NUM_HARDWARE)
	return -EINVAL;
    hardware[type] = desc;
    return 0;
}

int
unregister_hardware(hw_type type)
{
    if (type < 0 || type >= NUM_HARDWARE)
	return -EINVAL;
    hardware[type] = NULL;
    return 0;
}

#ifdef __linux__

/* try just a system call if the trap mechanism is busted */
#ifdef TRAP_SYSTEM_CALL
extern void (*sys_call_table[])(void);
static void (*old_syscall)(void) = 0;
#endif

#define VERSION(v,p,s) (((v)<<16)+(p<<8)+s)

#if LINUX_VERSION_CODE >= VERSION(2,1,121)
#define _set_gate(gate_addr,type,dpl,addr) \
__asm__ __volatile__ ("movw %%dx,%%ax\n\t" \
        "movw %2,%%dx\n\t" \
        "movl %%eax,%0\n\t" \
        "movl %%edx,%1" \
        :"=m" (*((long *) (gate_addr))), \
         "=m" (*(1+(long *) (gate_addr))) \
        :"i" ((short) (0x8000+(dpl<<13)+(type<<8))), \
         "d" ((char *) (addr)),"a" (__KERNEL_CS << 16) \
        :"ax","dx")

struct desc_struct { unsigned long a, b; };
extern struct desc_struct idt_table[];  /* arch/i386/kernel/traps.c only */

#define set_intr_gate(n,addr)   _set_gate(idt_table+(n),14,0,addr)
#define set_trap_gate(n,addr)   _set_gate(idt_table+(n),15,0,addr)
#define set_system_gate(n,addr) _set_gate(idt_table+(n),15,3,addr)
#define set_call_gate(a,addr)   _set_gate(a,12,3,addr)

/*
 * Here's the info on x86 trap sorts of things.
 *
 * The hardware supports 4 different Gate Descriptors which can transfer
 * between different privilege levels.
 *
 *   Call gates - live in LDT or GDT, not IDT.  Lots of functionality for
 *     calling a subroutine in a different level, including stack switching
 *     and copying of parameters from one stack to the next.
 *
 *   Trap gates - Live in IDT.  This is the normal thing.
 *
 *   Interrupt gates - Same as Trap gate, but IF is cleared on entry and
 *     set on exit (IRET); just like cli()/sti() wrapper around the Trap
 *     gate code.
 *
 *   Task gates - Live in IDT.  Initiates a task switch when the trap
 *     arrives and the vector in the IDT is of Task gate type.
 *
 * How linux uses these things:
 *   set_intr_gate() - Interrupt gate, cli()/sti() wrapped, DPL 0 only.
 *   set_trap_gate() - Trap gate, DPL 0.
 *   set_system_gate() - Trap gate, any DPL.
 *   set_call_gate() - Call gate, into LDT or GDT, any privilege.
 * As of 2.1.121, these are used as follows:
 *   set_intr_gate(): real-live external interrupts.
 *   set_trap_gate(): INT numbers 0-2,6-48.
 *   set_system_gate(): INT numbers 3 (breakpoint), 4 (overflow), 5 (bound),
 *     80 (system_call).
 *   set_call_gate(): default_ldt, used to setup personality.
 *
 * So we want set_system_gate() to let the user get at a kernel routine,
 * just like the normal system_call entry.
 */
#endif

static void
set_vi_trap_gate(void (*func)(void))
{
#ifdef TRAP_SYSTEM_CALL  /* try just a system call */
    if (func != 0) {
	old_syscall = sys_call_table[VI_GATE];
	sys_call_table[VI_GATE] =  func;  /* probably vi_service */
    } else {
	sys_call_table[VI_GATE] = old_syscall;
	old_syscall = 0;
    }
#else  /* TRAP_SYSTEM_CALL */
    set_system_gate(VI_GATE, &vi_trap);
#endif  /* TRAP_SYSTEM_CALL */
}
#endif  /* __linux__ */

/*
 * Called by main() during pseudo-device attachment.  All we need
 * to do is allocate enough space for devices to be configured later, and
 * add devsw entries.
 */
static void
vi_attach(void *dummy)
{
    int nvi = NVI;
#if defined(__FreeBSD__) && defined(DEVFS)
    int i;
#endif

    if (nvi > 1)
	printf("vi0-%d: Virtual Interfaces\n", nvi - 1);
    else
	printf("vi0: Virtual Interface\n");

    vi_softc = (struct vi_softc *)malloc(nvi *
	sizeof(struct vi_softc), M_DEVBUF, M_NOWAIT);
    if (vi_softc == NULL)
	goto memfail;
    bzero(vi_softc, sizeof(struct vi_softc) * nvi);

    vis = (struct vi_instance **)malloc(VI_MAXVI *
	sizeof(struct vi_instance *), M_DEVBUF, M_NOWAIT);
    if (vis == NULL)
	goto memfail;
    bzero(vis, sizeof(struct vi_instance *) * VI_MAXVI);

    /* We are committed now, the following must succeed. */
#ifdef __FreeBSD__
#ifdef DEVFS
    for (i = 0; i < nvi; i++) {
	vi_softc[i]->devfs_token = devfs_add_devswf(&vi_cdevsw, i, DV_CHR,
	    UID_ROOT, GID_WHEEL, 0600, "vi%d", i);
    }
#else
    vi_cdevhandle = makedev(cdev_major, 0);
    cdevsw_add(&vi_cdevhandle, &vi_cdevsw, NULL);
#endif
#endif  /* FreeBSD */
#ifdef __linux__
    for (;;) {
	if (register_chrdev(cdev_major, "vi", &vi_fops))
	    printk("vi: unable to get major %d, trying next\n", cdev_major);
	else
	    break;
	++cdev_major;
    }
#endif

    /* Set up call gate */
#ifdef __FreeBSD__
    setidt(VI_GATE, &IDTVEC(vi_trap), SDT_SYS386TGT, SEL_UPL,
	   GSEL(GCODE_SEL, SEL_KPL));
#endif
#ifdef __linux__
#  ifdef TRAP_SYSTEM_CALL
    set_vi_trap_gate((void (*)(void)) vi_service);  /* don't need the trap indirection */
#  else
    set_vi_trap_gate(vi_trap);
#  endif
#endif  /* __linux__ */

    return;

memfail:
    printf("vi: not enough memory for VI devices, bailing...\n");
    if (vi_softc)
	free(vi_softc, M_DEVBUF);
    if (vis)
	free(vis, M_DEVBUF);
}

/*
 * vi_open: Device open routine.
 */
static int
#ifdef __FreeBSD__
vi_open(dev_t dev, int flags, int fmt, struct proc *p)
#endif
#ifdef __linux__
vi_open(struct inode *inode, struct file *file)
#endif
{
#ifdef __FreeBSD__
    int unit = minor(dev);
#if VI_DEBUG > 1
    printf("vi%d: open: flags=%d, fmt=%d\n", unit, flags, fmt);
#endif
#endif
#ifdef __linux__
    unsigned int unit = MINOR(inode->i_rdev);
#endif

    vi_softc[unit].refcnt++;

    return 0;
}

/*
 * vi_close: Device close routine.  Make sure all of the resources that
 *	     the process held are freed properly.
 */
static int
#ifdef __FreeBSD__
vi_close(dev_t dev, int flags, int fmt, struct proc *p)
#endif
#ifdef __linux__
vi_close(struct inode *inode, struct file *file)
#endif
{
    int i;
#ifdef __FreeBSD__
    int unit = minor(dev);
#endif
#ifdef __linux__
    unsigned int unit = MINOR(inode->i_rdev);
    struct task_struct *p = current;
#endif
    struct vi_softc *sc = &vi_softc[unit];
    struct vi_instance *vi;
    struct vi_memhandle *mh;

#if VI_DEBUG > 2
#ifdef __FreeBSD__
    printf("vi%d: close: flags=%d, fmt=%d\n", unit, flags, fmt);
#endif
#endif

    for (i = 0; i < VI_MAXMH; i++) {
	mh = sc->mem_handles[i];

	if (mh && mh->owner == p) {
#if VI_DEBUG > 1
	    printf("vi_close: unmapping handle %d\n", i);
#endif
	    vi_unmapmh(mh);
	    free(mh, M_TEMP);
	    sc->mem_handles[i] = NULL;
	    sc->num_memreg--;
	}
    }

    for (i = 0; i < VI_MAXVI; i++) {
	vi = vis[i];
	if (vi && vi->owner == p) {
#if VI_DEBUG > 1
	    printf("vi_close: freeing vi %d\n", i);
#endif
	    free(vi, M_TEMP);
	    vis[i] = NULL;
	    num_vi--;
	}
    }

    sc->refcnt--;

    return 0;
}

/*
 * vi_ioctl: Used to handle various management of the kernel structures.
 */
int
#ifdef __FreeBSD__
vi_ioctl(dev_t dev, ioctlcmd cmd, caddr_t data, int flag, struct proc *p)
#endif
#ifdef __linux__
vi_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
  unsigned long arg)
#endif
{
    int error = 0;
#ifdef __FreeBSD__
    int unit = minor(dev);
#endif
#ifdef __linux__
    unsigned int unit = MINOR(inode->i_rdev);
#endif
    struct vi_softc *sc = &vi_softc[unit];

    switch (cmd) {
    case VI_IFATTACH: {
#ifdef __FreeBSD__
	struct ifnet *ifp;
#endif
#ifdef __linux__
	struct device *ifp;
	char data[16];
	int i;
#endif
	if (sc->ifp != NULL) {
	    error = EBUSY;
	    break;
	}
#ifdef __FreeBSD__
	ifp = ifunit((char *)data);
#endif
#ifdef __linux__
	if (copy_from_user(data, (void *)arg, sizeof(data)))
	    return -EFAULT;
	ifp = dev_get(data);
#endif
	if (ifp == NULL) {
	    error = EINVAL;
	    break;
	}

	/* Find a supporting hardware interface for the dev */
	for (i=0; i<NUM_HARDWARE; i++)
	    if (hardware[i] && hardware[i]->attach)
		if ((hardware[i]->attach)(ifp, vi_input) == 0) {
		    sc->hw = i;
		    sc->ifp = ifp;
		    break;
		}
	if (i == NUM_HARDWARE) {
	    error = ENODEV;
	    break;
	}
#if VI_DEBUG > 1
	printf("vi%d: Attaching to interface %s\n", unit, (char *)data);
#endif
	break;
    }

    case VI_IFDETACH:
	if (sc->ifp == NULL) {
	    error = EINVAL;
	    break;
	}
	if (sc->refcnt != 1) {
	    /* must be open by exactly the detaching process */
	    printk("Okay, detaching, but sc->refcnt = %d\n", sc->refcnt);
	    /* error = EBUSY; break; */
	}
	/*
	 * Check for open vi's, free resources
	 */
#if VI_DEBUG > 1
	printf("vi%d: Detaching from interface\n", unit);
#endif
	if (hardware[sc->hw]->detach)
	    (hardware[sc->hw]->detach)();
	sc->ifp = NULL;
	sc->hw = HW_NONE;
	break;

    case VI_INFO: {
#ifdef __FreeBSD__
	struct vi_info_req *req = (struct vi_info_req *)data;
#endif
#ifdef __linux__
	struct vi_info_req the_req, *req = &the_req;
	if (copy_from_user(req, (void *)arg, sizeof(struct vi_info_req)))
	    return -EFAULT;
#endif

	if (sc->ifp == NULL) {
	    error = EINVAL;
	    break;
	}
#ifdef __FreeBSD__
	strncpy(req->if_name, sc->ifp->if_name, 10);
	req->if_unit = sc->ifp->if_unit;
#endif
#ifdef __linux__
	strcpy(req->if_name, sc->ifp->name);
	req->if_unit = 0;  /* not relevant in linux */
	copy_to_user((void *)arg, req, sizeof(struct vi_info_req));
#endif
	break;
    }

    case VI_CREATE: {
	struct vi_instance *v;
#ifdef __FreeBSD__
	struct vi_create_req *req = (struct vi_create_req *)data;
	struct ether_header *eh;
#endif
#ifdef __linux__
	/* hope sizeof(vi_create_req) < sizeof(ifreq) */
	struct vi_create_req the_req, *req = &the_req;
	struct ethhdr *eh;
	copy_from_user(req, (void *)arg, sizeof(struct vi_create_req));
#endif

	if (num_vi == VI_MAXVI) {
	    error = EBUSY;
	    break;
	}
	if (sc->ifp == NULL) {
	    error = EINVAL;
	    break;
	}
	if (req->hw < 0 || req->hw >= NUM_HARDWARE || !hardware[req->hw]
	 || req->hw != sc->hw) {
	    error = ENODEV;
	    break;
	}
	req->ViHandle = VI_MALLOC(vis, sc->cur_vi, VI_MAXVI);
	v = (struct vi_instance *) malloc(
	    sizeof(struct vi_instance), M_TEMP, M_NOWAIT);
	if (v == NULL) {
	    error = ENOMEM;
	    break;
	}
	bzero(v, sizeof(struct vi_instance));
	vis[req->ViHandle] = v;

	v->hw = req->hw;
	v->ifp = sc->ifp;
#ifdef __FreeBSD__
	v->owner = p;
	v->mtu = sc->ifp->if_mtu - sizeof(struct vi_hdr);
#endif
#ifdef __linux__
	v->owner = current;
	v->mtu = sc->ifp->mtu - sizeof(struct vi_hdr);
#endif
	v->mem_handles = sc->mem_handles;	/* for speed */
	v->state = VI_IDLE;
	v->send_off = 0;
	v->vh_flags = 0;

	eh = (struct ethhdr *) &v->header_template;
	v->svi = req->ViHandle;

	/* call driver layer to fill in its stuff */
	if (hardware[v->hw]->create)
	    (hardware[v->hw]->create)(v, sc->ifp);

#ifdef __linux__
	/* send back the new ViHandle */
	copy_to_user((void *)arg, req, sizeof(struct vi_create_req));
#endif
	num_vi++;
	break;
    }

    case VI_DESTROY: {
	int vi_idx;
	struct vi_instance *v;

#ifdef __FreeBSD__
	vi_idx = *(int *)data;
#endif
#ifdef __linux__
	if (copy_from_user(&vi_idx, (void *)arg, sizeof(int)))
	    return -EFAULT;
#endif

	v = vis[vi_idx];
	if (!v)
	    return -EINVAL;

	if (hardware[v->hw]->destroy)
	    (hardware[v->hw]->destroy)(v);

	free(v, M_TEMP);
	vis[vi_idx] = NULL;

	num_vi--;
	break;
    }

    case VI_MEM_REG: {
	struct vi_memhandle *mh;
#ifdef __FreeBSD__
	struct vi_memreg_req *req = (struct vi_memreg_req *)data;
#endif
#ifdef __linux__
	struct task_struct *p = current;
	struct vi_memreg_req the_req, *req = &the_req;

	copy_from_user(req, (void *)arg, sizeof(struct vi_memreg_req));
#endif
	if (sc->num_memreg == VI_MAXMH) {
	    error = EBUSY;
	    break;
	}
	req->mem_handle = VI_MALLOC(sc->mem_handles, sc->cur_mh, VI_MAXMH);
	mh = (struct vi_memhandle *)malloc(sizeof(struct vi_memhandle),
	    M_TEMP, M_NOWAIT);
	if (mh == NULL) {
	    error = ENOMEM;
	    break;
	}

	mh->owner = p;
	mh->uaddr = req->addr;
	mh->len = req->len;
	if ((error = vi_mapmh(mh))) {
	    free(mh, M_TEMP);
	    break;
	}

	sc->mem_handles[req->mem_handle] = mh;
	sc->num_memreg++;
#ifdef __linux__
	/* return the result */
	copy_to_user((void *)arg, req, sizeof(struct vi_memreg_req));
#endif
	break;
    }

    case VI_MEM_DEREG: {
	struct vi_memhandle *mh;
#ifdef __FreeBSD__
	struct vi_memreg_req *req = (struct vi_memreg_req *)data;
#endif
#ifdef __linux__
	struct vi_memreg_req the_req, *req = &the_req;
	struct task_struct *p = current;

	if (copy_from_user(req, (void *)arg, sizeof(struct vi_memreg_req)))
	    return -EFAULT;
#endif

	mh = sc->mem_handles[req->mem_handle];
	if (mh == NULL || mh->owner != p) {
	    error = EINVAL;
	    break;
	}

	vi_unmapmh(mh);
	sc->mem_handles[req->mem_handle] = NULL;

	sc->num_memreg--;
	break;
    }

    case VI_DESC_WAIT: {
/*


	tsleep(...);
	break;
*/
    }

    case VI_BIND: {
	struct vi_instance *v;
#ifdef __FreeBSD__
	struct vi_modify_req *req = (struct vi_modify_req *)data;
	struct ether_header *eh;
#endif
#ifdef __linux__
	struct vi_bind_req the_req, *req = &the_req;
	copy_from_user(req, (void *)arg, sizeof(struct vi_bind_req));
#endif

	if (req->vi > VI_MAXVI) {
	    printk("vi_modify: vi handle %d out of range\n", req->vi);
	    break;
	}
	v = vis[req->vi];
	if (v == NULL) {
	    printf("vi_modify: null vi\n");
	    break;
	}
	if (hardware[v->hw]->bind)
	    error = (hardware[v->hw]->bind)(v, req, (struct vi_bind_req *)arg);
	break;
    }

    case VI_STATS: {
#ifdef __FreeBSD__
	struct vi_stats *req = (struct vi_stats *)data;
	*req = vi_stats;
#endif
#ifdef __linux__
	if (!access_ok(VERIFY_WRITE, (void *)arg, sizeof(struct vi_stats)))
	    return -EFAULT;
	copy_to_user((void *)arg, &vi_stats, sizeof(struct vi_stats));
#endif
	break;
    }

    case VI_GETPEER: {
	struct vi_getpeer_req req;
	struct vi_instance *v;

	copy_from_user(&req, (void *)arg, sizeof(struct vi_getpeer_req));
	if (req.vi > VI_MAXVI) {
	    printk("vi_modify: vi handle %d out of range\n", req.vi);
	    break;
	}
	v = vis[req.vi];
	if (v == NULL) {
	    error = EINVAL;
	    break;
	}
	if (hardware[v->hw]->bind)
	    error = (hardware[v->hw]->getpeer)(v, &req,
	      (struct vi_getpeer_req *)arg);
	break;
    }

    default:
	error = ENOTTY;
    }

#ifdef __FreeBSD__
    return error;
#endif
#ifdef __linux__
    return -error;
#endif
}

static void
hexdump(void *vwhere, int n)
{
    int i;
    unsigned char *where = (char *)vwhere;

    while (n > 0) {
	printk("%p ", where);
	for (i=0; i < ( (n>16)?16:n); i++)
	    printk(" %02x", where[i]);
	printk("\n");
	n -= 16;
	where += 16;
    }
}

/*
 * Utility functions for playing with memory.
 */
#ifdef __linux__
#define vm_offset_t off_t
#endif

/*
 * Use a memory handle to translate user addresses into kernel addresses.
 * Also, return NULL if the memory is not completely within the registered
 * handle.
 */
static inline caddr_t
vi_utok(struct vi_instance *vi, caddr_t user, int mh_idx, int len)
{
    struct vi_memhandle *mh;
    vm_offset_t off;

    mh = vi->mem_handles[mh_idx & VI_MHMASK];
    if (mh == NULL) {
	printf("vi_utok: null memory handle idx=%d.\n", mh_idx);
	return NULL;
    }
    off = user - mh->uaddr;

    if (off < 0 || off + len > mh->len) {
	printf("vi_utok: memory not within handle: "
	  "user %p into mh->uaddr %p, mh->len " LXORX "\n",
	  user, mh->uaddr, mh->len);
	return NULL;
    }
    return mh->kaddr + off;
}

/*
 * Check arguments and return the kernel-space version of the passed-in
 * user handle.
 */
static inline struct vi_cseg *
vi_desc(struct vi_cseg *cs, register struct vi_memhandle *mh)
{
    vm_offset_t off;
    vm_offset_t end;

    if (cs == NULL || mh == NULL) {
	printf("vi_desc: null %s.\n", cs ? "memory handle" : "descriptor");
	return NULL;
    }
    off = (caddr_t)cs - mh->uaddr;
    end = off + sizeof(struct vi_cseg);
    if (off < 0 || end > mh->len) {
	printf("vi_desc: cseg not within handle: "
	  "%p into mh->uaddr %p, mh->len " LXORX "\n", cs, mh->uaddr, mh->len);
	return NULL;
    }
    if (end + cs->SegCount * sizeof(struct vi_dseg) > mh->len) {
	cs->Status = VIP_STATUS_PARTIAL_ERROR;
	printf("vi_desc: partial descriptor.\n");
	return NULL;
    }
    return (struct vi_cseg *)(mh->kaddr + off);
}

static inline struct vi_cseg *
vi_next_desc(struct vi_instance *vi, struct vi_cseg *cs)
{
    if (cs == NULL || vi == NULL) {
	printf("vi_next_desc: null %s.\n", cs ? "vi" : "descriptor");
	return NULL;
    }

    if (cs->Next.Address)
	return vi_desc(cs->Next.Address,
	               vi->mem_handles[cs->NextHandle & VI_MHMASK]);
    /* not an error if there is no next address */
    else return 0;
}
#ifdef __linux__
#undef vm_offset_t
#endif

/*
 * vi_input(): Called from ether_input() with a contiguous data mbuf.
 *	       This routine copies the mbuf to the apropriate location
 *	       as specified by the next receive descriptor.  All
 *	       descriptors/data have already been verified.
 */
#ifdef __FreeBSD__
static void
vi_input(struct mbuf *m)
#endif
#ifdef __linux__
static void
vi_input(struct sk_buff *skb)
#endif
{
    struct vi_instance *vi = NULL;
    struct vi_hdr *vh;
    struct vi_cseg *cs;
    struct vi_dseg *ds;
    caddr_t data, kaddr;
    int len;

    /* printk("vi_input:\n"); hexdump(skb->data, skb->len); */
#ifdef __FreeBSD__
    if (m->m_len < sizeof(struct vi_hdr)) {
	printf("vi_input: mbuf too short.\n");
	goto free_return;
    }
    len = m->m_len - sizeof(struct vi_hdr);
    vh = mtod(m, struct vi_hdr *);
#endif
#ifdef __linux__
    if (skb->len < sizeof(struct vi_hdr)) {
	printf("vi_input: mbuf too short.\n");
	goto freereturn;
    }
    len = skb->len - sizeof(struct vi_hdr);
    vh = (struct vi_hdr *)skb->data;  /* pointer to start of "actual" data? */
#endif

#if VI_DEBUG > 1
    printf("vi_input: header: svi=%d, dvi=%d, len=%d, off=%d, flags=%x\n",
	   vh->svi, vh->dvi, vh->len, vh->off, vh->flags);
#endif

    if (vh->len > len) {
	printf("vi_input: packet too small: len is %d, should be %d.\n",
		len, vh->len);
	goto freereturn;
    } else if (vh->len < len)
	len = vh->len;

    if (vh->dvi & ~VI_VIMASK) {
	int i;
	/* connect request for things bigger than the registered range */
	for (i=0; i<VI_MAXVI; i++)
	    if (vis[i] && vis[i]->state == VI_CONNECTING
	               && vis[i]->dvi == vh->dvi) {
		vi = vis[i];
		break;
	    }
    } else {
	/* normal path */
	vi = vis[vh->dvi & VI_VIMASK];
    }
    data = (caddr_t)(vh + 1);

    if (vi == NULL) {
	printf("vi_input: NULL vi on dvi %d:\n", vh->dvi);
	printf("vi_input: header: svi=%d, dvi=%d, len=%d, off=%d, flags=%x\n",
	       vh->svi, vh->dvi, vh->len, vh->off, vh->flags);
	hexdump(skb->data, skb->len);
	goto freereturn;
    }

#if 0
    /* not sure if I want this to happen */
    if (vh->svi != vi->dvi) {
	printf("vi_input: mismatch: svi %d, dvi %d:\n", vh->svi, vi->dvi);
	hexdump(skb->data, skb->len);
	goto freereturn;
    }
#endif

    cs = vi->recvq.head;
    if (cs == NULL) {
	printf("vi_input: Packet received with no pre-post:\n");
	hexdump(skb->data, skb->len);
	goto freereturn;
    }

    if (vh->flags) {
	/* long path: this is not a normal data packet. */
	if (vh->flags == VI_CONNECT) {
	    /* remember who called */
	    if (hardware[vi->hw]->store_peer)
		(hardware[vi->hw]->store_peer)(vi, skb);
	} else
	    printk("vi_input: unknown flags %02x\n", vh->flags);
    }

    if (cs->Control & VIP_CONTROL_IMMEDIATE) {
	if (len != 4) {
	    printk("vi_input: IMMEDIATE with length %d, expected 4\n", len);
	    goto freereturn;
	}
	vi->recvq.head = vi_next_desc(vi, cs);
	cs->ImmediateData = *(u_int32_t *)data;
    } else {
	ds = (struct vi_dseg *)(cs + 1);	/* XXX this sucks. */
	if (ds->Length > vi->mtu) {
	    if (vh->off != vi->recv_off) {
		 printf("vi_input: out of order, got %d, wanted %d\n",
		   vh->off, vi->recv_off);
		 return;
	    }
	    kaddr = vi_utok(vi, (caddr_t)ds->Data.Address + vi->recv_off,
			    ds->Handle, len);
	    vi->recv_off += len;
	    if (vi->recv_off >= ds->Length)
		vi->recvq.head = vi_next_desc(vi, cs);
	} else if (ds->Length <= len) {
	    len = ds->Length;
	    kaddr = vi_utok(vi, ds->Data.Address, ds->Handle, len);
	    vi->recvq.head = vi_next_desc(vi, cs);
	} else {
	    printf("vi_input: not enough data for descriptor: ds=%d, len=%d\n",
		    ds->Length, len);
	    goto freereturn;
	}

	if (kaddr == NULL) {
	    printf("vi_input: bad descriptor data segment.\n");
	    goto freereturn;
	}
#ifdef __FreeBSD__
	bcopy(data, kaddr, len);
#endif
#ifdef __linux__
	memcpy(kaddr, data, len);
#endif
    }

    if (vi->recvq.head != cs) {
	cs->Status = VIP_STATUS_DONE;
	vi->recv_off = 0;
    }

freereturn:
#ifdef __FreeBSD__
    m_freem(m);
#endif
#ifdef __linux__
    kfree_skb(skb);
#endif
}

/*
 * Service send queue for VI.
 */
__inline void
vi_send(struct vi_instance *vi, struct vi_cseg *cs)
{
    int error;
#ifdef __FreeBSD__
    struct ifnet *ifp;
    struct mbuf *mb_head, *m;
    int cnt;
#endif

    /*
     * If we were passed a descriptor and the work queue is empty, add it
     * to the work queue.  Otherwise ignore the descriptor.  The idea is
     * that the user code will chain onto the end of the descriptor list
     * if it exists.
     */
    if (vi->sendq.head == NULL) {
	vi->sendq.head = cs;
    } else {
	cs = vi->sendq.head;
    }

    /*
     * Start sending from current descriptor.  At this point, there must
     * be at least one descriptor on the send work queue.
     */
    while (cs) {
#if VI_DEBUG > 1
#ifdef __FreeBSD__
	printf("vi_send: processing send descriptor %x\n", cs);
#endif
#ifdef __linux__
	printf("vi_send: processing send descriptor %p\n", cs);
#endif
#endif 
	if ((cs->Control & 3) == VIP_CONTROL_OP_SENDRECV) {
	    /* other control flags handled in driver */
	    /* printk("vi_send: to d_o(%p,%p), send_off = %d\n", cs, vi, vi->send_off); */
	    if ((error = (hardware[vi->hw]->transmit)(vi, cs)) == 0) {
		cs->Status = VIP_STATUS_DONE;
		vi->sendq.head = vi_next_desc(vi, vi->sendq.head);
	    } else if (error > 0) {
		continue;	/* not finished */
	    } else {
		/* write an error descriptor */
	    }
	} else {
	    printk("vi_send: Unsupported RDMA read/write, descr %p\n", cs);
	}

	cs = vi_next_desc(vi, cs);
    }
}

/*
 * Service receive queue for VI.
 */
__inline void
vi_recv(struct vi_instance *vi, struct vi_cseg *cs)
{
#ifdef __FreeBSD__
    int s;
#endif
#ifdef __linux__
    unsigned long flags;
#endif

#ifdef __FreeBSD__
    s = splimp();
#endif
#ifdef __linux__
    save_flags(flags); cli();
#endif
    if (vi->recvq.head == NULL) {
	vi->recvq.head = cs;
	vi->recvq.cur = cs;
    } else {
	cs = vi_next_desc(vi, vi->recvq.cur);
    }
#ifdef __FreeBSD__
    splx(s);
#endif
#ifdef __linux__
    restore_flags(flags);
#endif

    return;

    while (cs) {
#if VI_DEBUG > 1
#ifdef __FreeBSD__
	printf("vi_recv: processing send descriptor %x\n", cs);
#endif
#ifdef __linux__
	printf("vi_recv: processing send descriptor %p\n", cs);
#endif
#endif 
	switch (cs->Control) {
	case VIP_CONTROL_OP_SENDRECV:

	}
    }
}

void
#if defined(__linux__) && !defined(TRAP_SYSTEM_CALL)
vi_service(void)
#else
vi_service(struct vi_trap_args *args)
#endif
{
    struct vi_instance *vi;
    struct vi_cseg *cs;
    int index;

#if defined(__linux__) && !defined(TRAP_SYSTEM_CALL)
    struct vi_trap_args *args;  /* grab it from eax ourselves */
    asm volatile("movl %%eax, %0" : "=m" (args) :: "%eax");
#endif

    index = args->vi_idx & VI_VIMASK;
    if (index > VI_MAXVI) {
	printf("vi_service: vi index %d out of range\n", index);
	return;
    }

    vi = vis[index];
    if (vi == NULL) {
	printf("vi_service: null vi idx %d\n", args->vi_idx);
	return;
    }

    index = args->mh_idx & VI_MHMASK;
    if (index > VI_MAXMH) {
	printf("vi_service: mh index %d out of range\n", index);
	return;
    }
    cs = vi_desc(args->cs, vi->mem_handles[index]);
    /* printk("vi_service: vi_desc(%p, %p) = %p\n", args->cs, vi->mem_handles[index], cs); */
    /* printk("user:\n"); hexdump(args->cs, sizeof(struct vi_cseg)); */
    /* printk("krnl:\n"); hexdump(      cs, sizeof(struct vi_cseg)); */

    if (args->send)
	vi_send(vi, cs);
    else
	vi_recv(vi, cs);
}

/*
 * send descriptor free routine.  called by driver upon completion of
 * send for this mbuf.  also updates descriptor. assumes frees are done
 * in order.
 */
void
vi_free(caddr_t data, u_long int_vi)
{
    struct vi_instance *vi = (struct vi_instance *)int_vi;
    struct vi_cseg *cs;
#ifdef __FreeBSD__
    int s;
#endif
#ifdef __linux__
    unsigned long flags;
#endif
    
#ifdef __FreeBSD__
    s = splimp();
#endif
#ifdef __linux__
    save_flags(flags); cli();
#endif
    cs = vi->sendq.head;
/* XXX not yet.  this is a user addres.. how to convert now?
    ds = (struct vi_dseg *)(cs + 1);
    if (ds->Data.Address != data) {	
	printf("Out of order completion!\n");
	splx(s);
	return;
    }
*/
    cs->Status |= VIP_STATUS_DONE;
    vi->sendq.head = vi_next_desc(vi, cs);	/* XXX */
#ifdef __FreeBSD__
    splx(s);
#endif
#ifdef __linux__
    restore_flags(flags);
#endif

#if VI_DEBUG > 1
#ifdef __FreeBSD__
    printf("vi_free: done, new head=%x\n", vi->sendq.head);
#endif
#ifdef __linux__
    printf("vi_free: done, new head=%p\n", vi->sendq.head);
#endif
#endif 
}

#ifdef __linux__
/*
 * Linux memory mapping helper functions.
 *
 * Note:  remap_page_ranges() can be used to map a _physical_ address range
 * into the current process's virtual address space.  What I want to do is
 * to map each of a set of the current process's virtual pages into _kernel_
 * vmalloc area.  So:
 *
 * Confused.  If user says vi_map(..) of a five-page region which consists
 * of a two-page physically contiguous section, and another three-page
 * physically contiguous section, I can't produce a single physical address
 * from those.  I could produce a single kernel virtual address which has
 * the same mappings to those underlying physical pages, but I don't see
 * how this will make the device driver real happy when it thinks they're
 * all really physically contiguous.  How can Chris's stuff work if he
 * passes __pa(starting_kva) to the device, when actually the length would
 * cross out of the first two-page physically contiguous section?
 */

/*
 * This is the physical memory address of the start of a (user) page, offset
 * into page is ignored.
 */
unsigned long
kva_from_uaddr(unsigned long addr)
{
    pte_t *pte;
    unsigned long kva;

    /* find pa of page containing addr */
    pte = pte_offset(pmd_offset(pgd_offset(current->mm, addr), addr), addr);
    if (!pte_present(*pte))
	return -EINVAL;
    kva = pte_page(*pte);
    return kva;
}

/*
 * Split into two vmas at location "at".
 */
static int
split_vma(struct vm_area_struct *orig, unsigned long at)
{
    struct vm_area_struct *new;

    if (!(new = kmem_cache_alloc(vm_area_cachep, SLAB_KERNEL)))
	return -ENOMEM;
    new->vm_mm = orig->vm_mm;
    new->vm_start = at;
    new->vm_end = orig->vm_end;
    new->vm_page_prot = orig->vm_page_prot;
    new->vm_flags = orig->vm_flags;
    if (orig->vm_file) {
	printk("split_vma: not expecting a file mapping at %08lx\n",
	  orig->vm_start);
	return -EINVAL;
    }
    new->vm_file = 0;
    new->vm_next_share = 0;
    new->vm_pprev_share = 0;
    new->vm_ops = orig->vm_ops;
    new->vm_offset = 0;
    new->vm_pte = 0;
    new->vm_next = orig->vm_next;
    if (new->vm_next != NULL)
	new->vm_next->vm_pprev = &new->vm_next;
    new->vm_pprev = &orig->vm_next;

    orig->vm_end = at;
    orig->vm_next = new;
    return 0;
}

/*
 * Create separated vmas for the virtual address range and lock them down
 * to prevent page out.
 */
static int
split_and_lock(struct mm_struct *mm, unsigned long start, unsigned long end)
{
    struct vm_area_struct *vma;
    unsigned long addr;

    for (addr = start; addr < end; ) {
	vma = find_vma(mm, addr);
	if (!vma) {
	    printk("split_and_lock: none found for %08lx\n", addr);
	    return -EINVAL;
	}
	/* printk("split_and_lock: addr %08lx gives vma %08lx -- %08lx\n", addr,
	  vma->vm_start, vma->vm_end); */

	if (vma->vm_end > end) {
	    int ret = split_vma(vma, end);
	    if (ret)
		return ret;
	    ++mm->map_count;
	}

	vma->vm_flags |= VM_LOCKED;
	++mm->locked_vm;

	addr = vma->vm_end;
    }
    /* keep things clean, if possible */
    merge_segments(mm, start, end);
    return 0;
}


/*
 * The reverse of split_and_lock.
 */
static int
unsplit_and_unlock(struct mm_struct *mm, unsigned long start, unsigned long end)
{
    struct vm_area_struct *vma;
    unsigned long addr;

    for (addr = start; addr < end; ) {
	vma = find_vma(mm, addr);
	if (!vma) {
	    printk("unsplit_and_unlock: none found for %lx\n", addr);
	    return -EINVAL;
	}

	if (vma->vm_end > end) {
	    int ret = split_vma(vma, end);
	    if (ret)
		return ret;
	    ++mm->map_count;
	}

	if (!(vma->vm_flags | VM_LOCKED))
	    printk("unsplit_and_unlock: not locked already %lx--%lx\n",
	      vma->vm_start, vma->vm_end);
	else {
	    vma->vm_flags &= ~VM_LOCKED;
	    --mm->locked_vm;
	}

	addr = vma->vm_end;
    }
    /* mm/mmap.c, attempt to put back together what split_and_lock took apart */
    merge_segments(mm, start, end);
    return 0;
}

/*
 * Rewrite to handle multiple page mappings.
 */
int
manypage_do_vi_mapmh(struct vi_memhandle *mh)
{
    unsigned long addr, uaddr;
    unsigned long start, end;
    struct mm_struct *mm = current->mm;
    pgd_t *pgd, oldpgd;
    pmd_t *pmd;
    pte_t *pte;
    unsigned long *pas;
    int i, j, numpages, regions;
    int ret;

    /* XXX: go through other mapped areas and make sure there's
     * no overlap, else can't unLOCK properly */

    down(&current->mm->mmap_sem);
    start = (unsigned long) CADDR_FLOOR(mh->uaddr);
    end   = (unsigned long) CADDR_CEIL(mh->uaddr + mh->len);
    numpages = (end - start) >> PAGE_SHIFT;
    if (numpages == 0) {
	ret = 0;
	goto out;
    }

    split_and_lock(mm, start, end);

    /* bring them in:  loop over handle_mm_fault */
    /* printk("do_vi_mapmh: to make_pages_present\n"); */
    make_pages_present(start, end);

    /* make a list of all the PAs corresponding to the VAs */
    pas = (unsigned long *) kmalloc(numpages * sizeof(unsigned long),
      GFP_KERNEL);
    if (!pas) {
	ret = -ENOMEM;
	goto out;
    }
    for (i=0, addr = start; addr < end; addr += PAGE_SIZE, ++i)
	pas[i] = kva_from_uaddr(addr);

    /*
     * go through and see how many physical ranges that corresponds to;
     * usually the pages are completely disjoint so that this is a waste,
     * but maybe will do something clever later to aggregate the physical
     * pages (bigphysarea?)
     */
    regions = 1;
    for (i=1; i<numpages; i++)
	if (pas[i] != pas[i-1] + PAGE_SIZE)
	    ++regions;

#if 0
    printk("do_vi_mapmh: %08lx -- %08lx is %d pages, %d regions\n",
	start, end, numpages, regions);
    for (i=0; i<numpages; i++)
	printk("region %d: %08lx\n", i, pas[i]);
#endif

    /* allocate the kaddrs structure to store the u/k mappings */
    mh->kaddrs = (struct vi_memhandle_kaddrs *) kmalloc(
	(regions+1) * sizeof(struct vi_memhandle_kaddrs), GFP_KERNEL);
    if (!mh->kaddrs) {
	kfree(pas);
	ret = -ENOMEM;
	goto out;
    }
    addr = start;
    j = 0;  /* pas counter */
    for (i=0; i<regions; i++) {
	mh->kaddrs[i].uaddr = (caddr_t) addr;
	mh->kaddrs[i].kaddr = (caddr_t) pas[j];
	/* printk("kaddr %d: va %08lx, pa %08lx\n", i, addr, pas[j]); */
	do {
	    addr += PAGE_SIZE;
	    j++;
	} while (i < regions && pas[j] == pas[j-1] + PAGE_SIZE);
    }
    /* upper bounds, for looping */
    mh->kaddrs[regions].uaddr = (caddr_t) end;
    mh->kaddrs[regions].kaddr = (caddr_t) (pas[numpages-1] + PAGE_SIZE);
    kfree(pas);

    /*
     * Set up the ptes starting at area->addr so that they point
     * to the contiguous set of user pages we identified above.
     * Just like remap_page_range() except that instead of messing
     * with current->mm, we use init_mm, and we don't check to make
     * sure that the mapped range is above max_mapnr, but just go
     * ahead and set the pte.
     */
    mh->area = get_vm_area(numpages * PAGE_SIZE);
    if (!mh->area) {
	ret = -ENOMEM;
	goto kaddrs_out;
    }
    /* convenience variable */
    mh->kaddr = (void *) ((unsigned long)mh->area->addr
      + ((unsigned long)mh->uaddr & (PAGE_SIZE-1)));
    flush_cache_all();
    for (uaddr = start, addr = (unsigned long) mh->area->addr;
      uaddr < end;
      uaddr += PAGE_SIZE, addr += PAGE_SIZE) {
	pgd = pgd_offset_k(addr);
	oldpgd = *pgd;
	/* if the pmd doesn't exist, this allocs, else just picks an entry */
	pmd = pmd_alloc_kernel(pgd, addr);
	ret = -ENOMEM;
	if (!pmd)
	    goto kaddrs_out;
	pte = pte_alloc_kernel(pmd, addr);
	if (!pte)
	    goto kaddrs_out;
	/* printk("kaddr %lx, pgd %p -> %lx, ", addr, pgd, pgd->pgd); */
	/* printk("pte %p -> %lx => ", pte, pte->pte); */

	set_pte(pte,
	  mk_pte((unsigned long) utophys(mh, (caddr_t)uaddr), PAGE_KERNEL));
	/* printk("%lx\n", pte->pte); */
	/* Not sure why this code is here; maybe it can go?  Watch it. */
	if (pgd_val(oldpgd) != pgd_val(*pgd)) {
	    printk("do_vi_mapmh:  pgdir changed %lx --> %lx\n", oldpgd.pgd,
	      pgd->pgd);
	    set_pgdir(addr, *pgd);
	}
    }
    flush_tlb_all();
    ret = 0;
    goto out;

  kaddrs_out:
    kfree(mh->kaddrs);
  out:
    up(&current->mm->mmap_sem);
    return ret;
}
#endif __linux__  /* mapmh helper functions */

static int
vi_mapmh(struct vi_memhandle *mh)
{
#ifdef __FreeBSD__
    register caddr_t addr, v, kva;
    vm_offset_t pa;

#if VI_DEBUG > 1
    printf("vi_mapmh: uaddr = %lx, len 0x%x\n", mh->uaddr, mh->len);
#endif

    kva = (caddr_t)kmem_alloc_pageable(kernel_map, mh->len + PAGE_SIZE);
    if (kva == NULL)
	return ENOBUFS;

    for (v = kva, addr = (caddr_t)trunc_page(mh->uaddr);
	addr < mh->uaddr + mh->len;
	addr += PAGE_SIZE, v += PAGE_SIZE) {
	    /*
	     * Do the vm_fault if needed; do the copy-on-write thing
	     * when reading stuff off device into memory.
	     */
	    vm_fault_quick(addr, VM_PROT_READ|VM_PROT_WRITE);
	    pa = trunc_page(pmap_kextract((vm_offset_t) addr));
	    if (pa == 0)
		    panic("vmapbuf: page not present");
	    vm_page_hold(PHYS_TO_VM_PAGE(pa));	/* XXX Wire instead? */
	    pmap_kenter((vm_offset_t) v, pa);
    }
    mh->kaddr = kva + (((vm_offset_t) mh->uaddr) & PAGE_MASK);
    return 0;
#endif  /* __FreeBSD__ */

#ifdef __linux__
#if VI_DEBUG > 0
    printf("vi_mapmh(uaddr = %p, len = 0x" LXORX ")\n", mh->uaddr, mh->len);
#endif
    return manypage_do_vi_mapmh(mh);
#endif
}

static void
vi_unmapmh(struct vi_memhandle *mh)
{
#ifdef __FreeBSD__
    register caddr_t addr;
    vm_offset_t pa;

#if VI_DEBUG > 1
    printf("vi_unmapmh(kaddr = %lx, len = %d\n", mh->kaddr, mh->len);
#endif
    for (addr = (caddr_t)trunc_page(mh->kaddr);
	addr < mh->kaddr + mh->len;
	addr += PAGE_SIZE) {
	    pa = trunc_page(pmap_kextract((vm_offset_t) addr));
	    pmap_kremove((vm_offset_t) addr);
	    vm_page_unhold(PHYS_TO_VM_PAGE(pa));
    }
#endif

#ifdef __linux__
    /* Do the reverse of vi_mapmh:
     *   unLOCK pages
     *   coalesce vmas
     *   (don't mess with user's pte/pmd/pgd; he's still got the pages)
     *   (don't make_pages_UNpresent; they'll go away with swap, etc.)
     *   clear kernel pte mappings
     *   put the vmarea back
     *   kfree(kaddrs);
     */
    unsigned long addr;

#if VI_DEBUG > 0
    printf("vi_unmapmh(uaddr = %p, len = " LXORX ")\n", mh->uaddr, mh->len);
#endif
    flush_cache_all();

    down(&current->mm->mmap_sem);
    /*
     * Unlock the pages as seen from the user, putting his vmas back in order.
     * Only do this if process isn't in the process of exiting, since otherwise
     * the mm has just been thrown away.
     */
    if (current->mm != &init_mm)
	unsplit_and_unlock(current->mm, (unsigned long) CADDR_FLOOR(mh->uaddr),
	  (unsigned long) CADDR_CEIL(mh->uaddr + mh->len));

    /* Delete the kernel mappings of the user pages */
    for (addr = (unsigned long) mh->area->addr;
      addr < (unsigned long) mh->area->addr + mh->area->size;
      addr += PAGE_SIZE) {
	pgd_t *pgd;
	pmd_t *pmd;
	pte_t *pte;

	/* These all better exist since we did map them in. */
	pgd = pgd_offset_k(addr);
	pmd = pmd_offset(pgd, addr);
	pte = pte_offset(pmd, addr);
	pte_clear(pte);
	/* but don't free the page; user process will get it */

	/*
	 * Also don't release any pmd/pgd we may have created, as
	 * they should be small (famous last words).  If you want to,
	 * it would mean running through all the ptes to make sure they're
	 * clear, then freeing the page.
	 */
	/* pte_free_kernel(pte); pmd_free_kernel(pmd); pgd_free(pgd); */
    }
    up(&current->mm->mmap_sem);
    flush_tlb_all();
    put_vm_area(mh->area);
    kfree(mh->kaddrs);
#endif
}

#ifdef __FreeBSD__
#ifndef VI_MODULE

PSEUDO_SET(vi_attach, vi);

#else

/*
 * Loadable VI pseudo-device stubs
 */
#include <sys/exec.h>
#include <sys/sysent.h>
#include <sys/lkm.h>

MOD_DEV(vi, LM_DT_CHAR, CDEV_MAJOR, &vi_cdevsw);

static int
vi_load(struct lkm_table *lkmtp, int cmd)
{

    vi_attach(NULL);

    return 0;
}

static int
vi_unload(struct lkm_table *lkmtp, int cmd)
{
    int i;

    for (i = 0; i < NVI; i++)
	if (num_vi != 0)	/* XXX use file opens instead */
	    return EBUSY;
    free(vi_softc, M_DEVBUF);

#ifdef DEVFS
    /* use devfs_remove_dev() to remove devfs nodes XXX */
#else
    cdevsw_add(&vi_cdevhandle, NULL, NULL);
#endif

    setidt(VI_GATE, &IDTVEC(rsvd), SDT_SYS386TGT, SEL_KPL,
	   GSEL(GCODE_SEL, SEL_KPL));

    via_input = NULL;
    return 0;
}

static int
vi_stat(struct lkm_table *lkmtp, int cmd)
{
    return 0;
}

int
vi_mod(struct lkm_table *lkmtp, int cmd, int ver)
{
#if __FreeBSD__ > 2
        DISPATCH(vi, lkmtp, cmd, ver, vi_load, vi_unload, vi_stat);
#else
        DISPATCH(lkmtp, cmd, ver, vi_load, vi_unload, vi_stat);
#endif
}

#endif /* VI_MODULE */
#endif /* FreeBSD initialization routines */

#ifdef __linux__
#ifdef MODULE

/* pass-in arguments here, if any */

int
init_module(void)
{
    vi_attach(NULL);
    return 0;
}

void
cleanup_module(void)
{
    if (num_vi != 0)  /* XXX use file opens instead */
	printk("vi: cleanup_module: num_vi = %d\n", num_vi);

    kfree(vi_softc);
    kfree(vis);

    /* call detach a lot? on each vi */

    printk("vi: cleanup_module: trap gate set to NULL\n");
    /* what's the reverse of this? */
    set_vi_trap_gate(0);

    unregister_chrdev(cdev_major, "vi");
    return;
}

#endif  /* MODULE */
#endif  /* linux initialization routines */

#endif /* (FreeBSD && NVI > 0)  ||  linux */

/* vim: set tags+=/usr/src/linux/tags: */
