/*
 * linux VI trap gate routine
 * $Id: vi_trap_linux_i386.S,v 1.1 1998/09/12 00:32:19 pw Exp $
 * Copyright (C) 1998 Chris Csanady and Pete Wyckoff
 */
#include <linux/linkage.h>
#include <asm/segment.h>

/*
 * A stack switch has occurred on the move into this higher privilege level.
 * On the stack is calling:  SS,ESP,EFLAGS,CS,EIP, but you don't need to mess
 * with any of those.
 *
 * Overhead of linux system_call vs. this routine involves the following:
 *   extra push/pop of orig_eax
 *   GET_CURRENT into ebx
 *   check for PF_TRACESYS
 *   lookup into sys_call_table
 *   (call to vi_service)
 *   bh_mask & bh_active --> possible do_bottom_half
 *   possible reschedule (if current->need_resched)
 *   possible signal delivery (if current->sigpending)
 */
ENTRY(vi_trap)
	/* save all */
	pushl %es
	pushl %ds

	pushl %eax

	pushl %ebp
	pushl %edi
	pushl %esi

	pushl %edx
	pushl %ecx
	pushl %ebx

	/* set up kernel segments */
	movl $(__KERNEL_DS),%edx
	movl %edx,%ds
	movl %edx,%es

	/* call vi_service which will get its arg from eax */
	call SYMBOL_NAME(vi_service)

	/* restore all */
	popl %ebx
	popl %ecx
	popl %edx

	popl %esi
	popl %edi
	popl %ebp

	popl %eax

	popl %ds
	popl %es

	iret
	ALIGN

