/*
 * $Id: vivar.h,v 1.17 1998/09/22 22:00:32 pw Exp $
 * Copyright (C) 1998 Chris Csanady and Pete Wyckoff
 */
#ifndef _LKM_VIVAR_H
#define _LKM_VIVAR_H

#if defined(__linux__) && !defined(__KERNEL__)
#include <sys/types.h>
#endif

#ifdef __FreeBSD__
#include <sys/queue.h>
#include <sys/time.h>

#if __FreeBSD__ > 2
typedef u_long ioctlcmd;
#else
typedef int ioctlcmd;
#endif
#endif  /* __FreeBSD__ */

typedef enum { false, true } bool;

/* Defaults for VIP */
#define VI_VISHIFT	8
#define VI_MHSHIFT	8
#define VI_CQSHIFT	6

#define VI_VIMASK	((1 << VI_VISHIFT) - 1)
#define VI_MAXVI	(1 << VI_VISHIFT)
#define VI_MHMASK	((1 << VI_MHSHIFT) - 1)
#define VI_MAXMH	(1 << VI_MHSHIFT)
#define VI_CQMASK	((1 << VI_CQSHIFT) - 1)
#define VI_MAXCQ	(1 << VI_CQSHIFT)

/*									*
 * VIP Specific data structurs and defines.				*
 *									*/

/* VI Descriptor->Control bits */
#define VIP_CONTROL_OP_SENDRECV		0x0000
#define VIP_CONTROL_OP_RDMAWRITE	0x0001
#define VIP_CONTROL_OP_RDMAREAD		0x0002
#define VIP_CONTROL_IMMEDIATE		0x0004
#define VIP_CONTROL_QFENCE		0x0008
/* attempting a connection */
#define VIP_CONTROL_CONNECT		0x0010

/* VI Descriptor->Status bits */
#define VIP_STATUS_DONE			0x00000001
#define VIP_STATUS_FORMAT_ERROR		0x00000002
#define VIP_STATUS_PROTECTION_ERROR	0x00000004
#define VIP_STATUS_LENGTH_ERROR		0x00000008
#define VIP_STATUS_PARTIAL_ERROR	0x00000010
#define VIP_STATUS_DESC_FLUSHED_ERROR	0x00000020
#define VIP_STATUS_TRANSPORT_ERROR	0x00000040
#define VIP_STATUS_RDMA_PROT_ERROR	0x00000080
#define VIP_STATUS_REMOTE_DESC_ERROR	0x00000100
#define VIP_STATUS_ERROR_MASK		0x000001FE
#define VIP_STATUS_OP_SEND		0x00000000
#define VIP_STATUS_OP_RECEIVE		0x00010000
#define VIP_STATUS_OP_RDMA_WRITE	0x00020000
#define VIP_STATUS_OP_REMOTE_RDMA_WRITE	0x00030000
#define VIP_STATUS_OP_RDMA_READ		0x00040000
#define VIP_STATUS_OP_MASK		0x00070000
#define VIP_STATUS_IMMEDIATE		0x00080000
#define VIP_STATUS_RESERVED		0xFFF0FD00

/* VIP Return values */
typedef enum {
    VIP_SUCCESS,
    VIP_NOT_DONE,
    VIP_INVALID_PARAMETER,
    VIP_ERROR_RESOURCE,
    VIP_TIMEOUT,
    VIP_REJECTED,
    VIP_INVALID_RELIABILITY_LEVEL,
    VIP_INVALID_MTU,
    VIP_INVALID_QOS,
    VIP_INVALID_PTAG,
    VIP_INVALID_RDMAREAD,
} VIP_RETURN;

/* VI states */
enum vi_state {
    VI_IDLE, VI_CONNECTED, VI_CONNECTING, VI_ERROR
};

typedef union {
    u_int64_t		AddressBits;
    void		*Address;
} vi_void64;

enum vi_reliability {
    VIP_SERVICE_UNRELIABLE,
    VIP_SERVICE_RELIABLE_DELIVERY,
    VIP_SERVICE_RELIABLE_RECEPTION
};

struct vi_attribs {
    int		ReliabilityLevel;
    u_long	MaxTransferSize;
    int		QoS;
    int		Ptag;
    bool	EnableRdmaWrite;
    bool	EnableRdmaRead;
};

struct vi_nic_attribs {

};

struct vi_mem_attr {

};

/*									*
 * Various VI structures.  Some of these are shared with the user.	*
 *									*/
#define VI_DATA		0x00
#define VI_ACK		0x01
#define VI_CONNECT	0x02
struct vi_hdr {
    u_int16_t			svi;	/* source vi */
    u_int16_t			dvi;	/* destination vi / connect port */
    u_int32_t			off;	/* ack too */
    u_int16_t			len;	/* packet length */
    char			flags;
};
/* sizeof vi_hdr = 12 */

struct vi_memhandle_kaddrs {  /* valid in kernel only */
    caddr_t kaddr;
    caddr_t uaddr;
};

struct vi_memhandle {
#ifdef __linux__
    struct task_struct		*owner;
#else
    struct proc			*owner;
#endif
    struct vi_memhandle_kaddrs	*kaddrs;  /* kernel array of k/u mappings */
    caddr_t			kaddr;    /* non-aligned mapping of uaddr */
    caddr_t			uaddr;
    size_t			len;
    struct vm_struct		*area;   /* so kernel can free it later */
};

struct vi_workq {
    struct vi_cseg		*head;
    struct vi_cseg		*cur;
    struct vi_cseg		*tail;
    int				size;	/* outstanding workitems */
    struct vi_compq		*compq;
};

struct vi_compq {
    void		*compq_base;
    void		*cur_desc;
    int			compq_size;
};

/*
 * Descriptor structures
 */
struct vi_cseg {
    vi_void64		Next;
    u_int32_t		NextHandle;
    u_int16_t		SegCount;
    u_int16_t		Control;
    u_int32_t		Reserved;
    u_int32_t		ImmediateData;
    u_int32_t		Length;
    u_int32_t		Status;
};

struct vi_address_seg {
    vi_void64		Data;
    u_int32_t		Handle;
    u_int32_t		Reserved;
};

struct vi_dseg {
    vi_void64		Data;
    u_int32_t		Handle;
    u_int32_t		Length;
};

struct vi_net_address {
    u_int16_t		HostAddressLen;
    u_int16_t		DiscriminatorLen;
    u_int8_t		HostAddress[1];
};

typedef enum { HW_NONE, HW_ETHER } hw_type;

#define VI_GATE		0x79	/* trap number */

#if defined(KERNEL) || defined(__KERNEL__)
#ifdef __FreeBSD__
#include <sys/ioccom.h>
#endif
#define VI_KVA_CHUNKS	32	/* # of KVA chunks */
#define VI_KVA_SIZE	32768	/* KVA chunk size */
#define VI_DESC_ALIGN	0x3F	/* align to 64 bytes */

/*
 * Hardware interface layer description
 */
struct vi_instance;
struct vi_bind_req;
struct vi_getpeer_req;
struct hardware_desc {
    int (*attach)(struct device *dev, void (*func)(struct sk_buff *skb));
    void (*detach)(void);
    void (*create)(struct vi_instance *v, struct device *dev);
    void (*destroy)(struct vi_instance *v);
    int (*bind)(struct vi_instance *v, struct vi_bind_req *req,
      struct vi_bind_req *userreq);
    int (*getpeer)(struct vi_instance *v, struct vi_getpeer_req *req,
      struct vi_getpeer_req *userreq);
    void (*store_peer)(struct vi_instance *v, struct sk_buff *skb);
    int (*transmit)(struct vi_instance *v, struct vi_cseg *cs);
};

/* defined in vi.c, for use by hardware modules */
extern int register_hardware(hw_type type, struct hardware_desc *desc);
extern int unregister_hardware(hw_type type);

struct vi_instance {
#ifdef __linux__
    struct task_struct		*owner;
    struct device		*ifp;
#else
    struct proc			*owner;
    struct ifnet		*ifp;
#endif
    int				mtu;
    enum vi_state		state;
    struct vi_workq		sendq;
    struct vi_workq		recvq;
    struct vi_attribs		attribs;
    struct vi_memhandle		**mem_handles;
    u_int16_t			svi;  /* same as index into vis[] now */
    u_int16_t			dvi;  /* vis[] index of peer */
    u_int32_t			send_off;	/* for SAR */
    u_int32_t			recv_off;
    char			vh_flags;
    hw_type			hw;
    /* maintained by hardware interface layer */
    struct vi_net_address 	*peer_addr;
    void			*header_template;
};

/* 
 * A virtual interface is described after initialization by this structure.
 */
struct vi_softc {
#ifdef __linux__
    struct device	*ifp;		/* pointer to network interface */
#else
    struct ifnet	*ifp;		/* pointer to network interface */
#endif
    int			num_vi;		/* number of open VI's */
    int			num_memreg;	/* number of memory areas reg'd */
    int			cur_vi;
    struct vi_compq	*comp_queues[VI_MAXCQ];
    int			cur_cq;
    struct vi_memhandle *mem_handles[VI_MAXMH];
    int			cur_mh;
    int			refcnt;
#ifdef DEVFS
    void		*devfs_token;
#endif
    hw_type		hw;
};

#if defined(__linux__) && defined(__KERNEL__)
#define CADDR_FLOOR(addr) (caddr_t) (((unsigned long)(addr)) & PAGE_MASK)
#define CADDR_CEIL(addr)  (caddr_t) PAGE_ALIGN((unsigned long)(addr))

static inline caddr_t
utophys(struct vi_memhandle *mh, caddr_t uaddr)
{
    struct vi_memhandle_kaddrs *kaddrs = mh->kaddrs;

    if (uaddr < CADDR_FLOOR(mh->uaddr)
      || uaddr >= CADDR_CEIL(mh->uaddr + mh->len)) {
	printk("utok: uaddr out of range %p\n", uaddr);
	return 0;
    }
    for (kaddrs = mh->kaddrs; ; kaddrs++)
	if (kaddrs->uaddr <= uaddr && uaddr < (kaddrs+1)->uaddr) break;
    return kaddrs->kaddr;
}
#endif

#endif /* KERNEL || __KERNEL__ */

/*
 * Statistics gathering.
 */
struct vi_stats {
    int			invalid;
    int			dropped;
    int			toosmall;
};

/*
 * Generic array allocator.  Arguments are the array (ent), the last
 * allocated entry (hint), and size of the array (max).  The return
 * value is the index into the array of a free element.  Do *NOT*
 * call this if the array is full.  (Oops, this is not ANSI)
 */
#define VI_MALLOC(ent,hint,max) ({					\
    while ((ent)[(hint) % (max)])					\
	(hint)++;							\
    (hint) %= (max);							\
})

/*
 * User interface into kernel.  Trap routines provide very low overhead,
 * used in processing new Tx/Rx descriptors.
 */
struct vi_trap_args {
    struct vi_cseg	*cs;
    int			vi_idx;
    int			mh_idx;
    char		send;
};

#ifdef __i386__
#ifdef TRAP_SYSTEM_CALL
#  define VI_TRAP(arg) do { \
    long __res; \
    asm volatile ("int $0x80" \
	: "=a" (__res) \
	: "0" (VI_GATE), "b" ((long)(arg))); \
} while (0)
#else
#  define VI_TRAP(arg) do {				\
    struct vi_trap_args * __ta = (arg);			\
    asm volatile ("movl %0, %%eax; int $0x79\n"		\
		  :					\
		  : "g" (__ta)				\
		  : "%eax");				\
} while (0)
#endif
#endif  /* __i386__ */

#ifdef __alpha__
#ifdef TRAP_SYSTEM_CALL
#  define VI_TRAP(arg) do { \
     register unsigned long v0 __asm__("$0"); \
     register unsigned long a0 __asm__("$16"); \
     register unsigned long a3 __asm__("$19"); \
     v0 = VI_GATE; \
     a0 = (unsigned long)(arg); \
     __asm__("callsys # %0 %1 %2 %3" \
       : "=r" (v0), "=r" (a3) \
       : "0" (v0), "r" (a0) \
       : "$1", "$2", "$3", "$4", "$5", "$6", "$7", "$8", \
         "$22", "$23", "$24", "$25", "$27", "$28"); \
   } while (0)
#else
#  define VI_TRAP(arg) do { printf("VI_TRAP undefined\n"); } while (0)
#endif
#endif  /* __alpha__ */

/*
 * ioctl() request structures
 */
struct vi_info_req {
    char		if_name[10];
    char		if_unit;
    int			data;
};

struct vi_create_req {
    struct vi_attribs	ViAttribs;
    int			SendCQHandle;
    int			RecvCQHandle;
    int			ViHandle;
    hw_type		hw;
};

struct vi_memreg_req {
    int			nic_handle;
    caddr_t		addr;
    size_t		len;
    struct vi_mem_attr	mem_attribs;
    int			mem_handle;
};

struct vi_test_req {
    int			vi;
    void		*data;
    int			size;
};

struct vi_bind_req {
    int vi;
    struct vi_net_address addr;
};

struct vi_getpeer_req {
    int vi;
    struct vi_net_address addr;  /* only Host part used */
};

/*
 * VI ioctl() calls
 */
#define VI_IFATTACH	 _IOW('F', 0, char [10])
#define VI_IFDETACH	  _IO('F', 1)
#define VI_INFO		 _IOR('F', 2, struct vi_info_req)
#define VI_CREATE	_IOWR('F', 3, struct vi_create_req)
#define VI_DESTROY	 _IOW('F', 4, int)
#define VI_MEM_REG	_IOWR('F', 5, struct vi_memreg_req)
#define VI_MEM_DEREG	_IOWR('F', 6, struct vi_memreg_req)
#define VI_DESC_WAIT	_IOWR('F', 7, int)
#define VI_BIND		 _IOW('F', 8, struct vi_bind_req)
#define VI_STATS	 _IOR('F', 9, struct vi_stats)
#define VI_GETPEER	 _IOR('F',10, struct vi_getpeer_req)

#endif  /* _LKM_VIVAR_H */
