/*
 * Header file of MP lite function declarations, as seen by an application
 * program.
 *   Pete Wyckoff
 */
#ifndef _MPFUNCS_H
#define _MPFUNCS_H

/* tcp.c */
extern void MP_Init(int *argc, char ***argv);
extern void MP_Finalize(void);
extern void MP_ASend(void *buf, int nbytes, int dest, int *msg_id);
extern void MP_Send(void *buf, int nbytes, int dest);
extern void MP_ARecv(void *buf, int nbytes, int src, int *msg_id);
extern void MP_Recv(void *buf, int nbytes, int src);
extern void MP_Wait(int *msg_id);
extern int MP_Nprocs(int *nprocs);
extern int MP_Myproc(int *myproc);
extern void MP_Sync(void);

#define DIF_SR(init,type,func) \
  extern inline void MP_ ## init ## func(type *buf, int count, int dest) \
    { MP_ ## func(buf, count * sizeof(type), dest); }
#define DIF_ASR(init,type,func) \
  extern inline void MP_ ## init ## func(type *buf, int count, int dest, int *msg_id) \
    { MP_ ## func(buf, count * sizeof(type), dest, msg_id); }
#define DIF_OP(init,type,func) \
  extern inline void MP_ ## init ## func(type *buf, int count) \
    { MP_ ## func(buf, count * sizeof(type)); }

/* mplite.c */
extern double MP_Time(void);
DIF_SR(d,double,Send) DIF_SR(i,int,Send) DIF_SR(f,float,Send)
DIF_SR(d,double,Recv) DIF_SR(i,int,Recv) DIF_SR(f,float,Recv)
DIF_ASR(d,double,ASend) DIF_ASR(i,int,ASend) DIF_ASR(f,float,ASend)
DIF_ASR(d,double,ARecv) DIF_ASR(i,int,ARecv) DIF_ASR(f,float,ARecv)
extern void MP_Broadcast(void *ptr, int nbytes);
DIF_OP(d,double,Broadcast); DIF_OP(i,int,Broadcast); DIF_OP(f,float,Broadcast);
extern void MP_Global(void *vec, int n, int op_type);
extern inline void MP_dSum(double *ptr, int n) { MP_Global(ptr, n, 0); }
extern inline void MP_iSum(int    *ptr, int n) { MP_Global(ptr, n, 1); }
extern inline void MP_fSum(float  *ptr, int n) { MP_Global(ptr, n, 2); }
extern inline void MP_dMax(double *ptr, int n) { MP_Global(ptr, n, 3); }
extern inline void MP_iMax(int    *ptr, int n) { MP_Global(ptr, n, 4); }
extern inline void MP_fMax(float  *ptr, int n) { MP_Global(ptr, n, 5); }
extern inline void MP_dMin(double *ptr, int n) { MP_Global(ptr, n, 6); }
extern inline void MP_iMin(int    *ptr, int n) { MP_Global(ptr, n, 7); }
extern inline void MP_fMin(float  *ptr, int n) { MP_Global(ptr, n, 8); }

#endif /* _MPFUNCS_H */
