/* Run mplite on top of the resident MPI implementation instead of TCP or SHMEM
 *   - Won't work for p4pg runs, since they use argc and argv arguments
 *     to MPI_Init() to start the processes on other nodes.  But this is why
 *     p4pg files don't work with Fortran codes at all anyway.
 * 
 */

#include "mplite.h"

#include <mpi.h>

static int mid = 0;
static void *msg_list[2000];

int MP_Myproc(int *rank)
{ MPI_Comm_rank(MPI_COMM_WORLD, rank); return *rank; }

int MP_Nprocs(int *size)
{ MPI_Comm_size(MPI_COMM_WORLD, size); return *size; }
 
void MP_Init(int *argc, char ***argv)
{
  if( *argc == 0 ) {         /* Assume this came from a Fortran program */
/*     printf("Using Fortran version of MPI_Init()\n");*/

#if defined (_CRAYMPP)
      MP_INIT_MPIF();
#elif defined (_AIX)
      mp_init_mpif();
#elif defined (sgi) || (__PARAGON__)
      mp_init_mpif_();
#else   /* Linux + ??? */
      mp_init_mpif__();
#endif

   } else {
/*      printf("Using C version of MPI_Init()\n");*/
      MPI_Init(argc, argv);
   }

   MP_Myproc(&myproc);
   MP_Nprocs(&nprocs);
   sprintf(logfilename, ".mplite.%d", myproc);
   if ( !(mylog=fopen(logfilename,"w")) ) {
     fprintf(stderr,"ERROR - Could not open log file %s\n",logfilename);
     exit(1);
   }
   fprintf(mylog,"Node %d of %d in MP_Init()\n",myproc,nprocs);fflush(mylog);
}

void MP_Finalize() { MPI_Finalize(); }

void MP_Send(void *buf, int nbytes, int dest)
{  MPI_Send(buf, nbytes, MPI_BYTE, dest, 0, MPI_COMM_WORLD); }

void MP_ASend(void *buf, int nbytes, int dest, int *msg_id)
{
   MPI_Request mpi_id;

   while( msg_list[++mid%2000] != NULL ) {};
   *msg_id = mid;

   MPI_Isend(buf, nbytes, MPI_BYTE, dest, 0, MPI_COMM_WORLD, &mpi_id);
   msg_list[*msg_id] = &mpi_id;
}

void MP_Recv(void *buf, int nbytes, int src)
{
   MPI_Status status;
   MPI_Recv(buf, nbytes, MPI_BYTE, src, 0, MPI_COMM_WORLD, &status);
}

void MP_ARecv(void *buf, int nbytes, int src, int *msg_id)
{
   MPI_Request mpi_id;

   while( msg_list[++mid%2000] != NULL ) {};
   *msg_id = mid;

   MPI_Irecv(buf, nbytes, MPI_BYTE, src, 0, MPI_COMM_WORLD, &mpi_id);
   msg_list[*msg_id] = &mpi_id;
}

void MP_Wait(int *msg_id)
{
   MPI_Status status;
   MPI_Wait((MPI_Request *) msg_list[*msg_id], &status);
   msg_list[*msg_id] = NULL;
}

void MP_Sync()
{ MPI_Barrier(MPI_COMM_WORLD); }

/*double MP_Time()*/
/*{ extern double MPI_Wtime(); return MPI_Wtime(); }*/


