/*
 * mplite.c
 * heavily modified by Pete Wyckoff, 1998.
 */
#define MPLITE_MAIN
#include "mplite.h"
#undef MPLITE_MAIN
#include "mpfuncs.h"

/*
 * gettimeofday() provides decent resolution (~1 ms at worst).
 * getrusage() counts CPU time instead, which is a problem since it
 * doesn't count the idle time while waiting on a packet.
 */
double
MP_Time(void)
{
    struct timeval f;

    gettimeofday(&f, (struct timezone *)NULL);
    return (double)f.tv_sec + (double)f.tv_usec * 1.0e-6;
}

#if 0
/*
 * Type-specific versions of byte-based functions
 */
void MP_dSend(void *buf, int count, int dest)
   { MP_Send( buf, count*sizeof(double), dest); }
void MP_iSend(void *buf, int count, int dest)
   { MP_Send( buf, count*sizeof(int), dest); }
void MP_fSend(void *buf, int count, int dest)
   { MP_Send( buf, count*sizeof(float), dest); }

void MP_dRecv(void *buf, int count, int source)
   { MP_Recv( buf, count*sizeof(double), source); }
void MP_iRecv(void *buf, int count, int source)
   { MP_Recv( buf, count*sizeof(int), source); }
void MP_fRecv(void *buf, int count, int source)
   { MP_Recv( buf, count*sizeof(float), source); }

void MP_dASend(void *buf, int count, int dest, int *msg_id)
   { MP_ASend( buf, count*sizeof(double), dest, msg_id); }
void MP_iASend(void *buf, int count, int dest, int *msg_id)
   { MP_ASend( buf, count*sizeof(int),    dest, msg_id); }
void MP_fASend(void *buf, int count, int dest, int *msg_id)
   { MP_ASend( buf, count*sizeof(float),  dest, msg_id); }

void MP_dARecv(void *buf, int count, int source, int *msg_id)
   { MP_ARecv( buf, count*sizeof(double), source, msg_id); }
void MP_iARecv(void *buf, int count, int source, int *msg_id)
   { MP_ARecv( buf, count*sizeof(int), source, msg_id); }
void MP_fARecv(void *buf, int count, int source, int *msg_id)
   { MP_ARecv( buf, count*sizeof(float), source, msg_id); }
#endif

/*
 * Broadcast from 0 to all other nodes
 */
void
MP_Broadcast(void *ptr, int nbytes)
{
    int dest;

    dmp3("Broadcast %d bytes at %p\n", nbytes, ptr);
    if (myproc == 0) {
        for (dest=1; dest<nprocs; dest++)
	    MP_Send(ptr, nbytes, dest);
    } else
        MP_Recv(ptr, nbytes, 0);
}

#if 0
void MP_dBroadcast(void *ptr, int count)
   { MP_Broadcast(ptr, count * sizeof(double) ); }
void MP_iBroadcast(void *ptr, int count)
   { MP_Broadcast(ptr, count * sizeof(int) ); }
void MP_fBroadcast(void *ptr, int count)
   { MP_Broadcast(ptr, count * sizeof(float) ); }
#endif


/*
 * Global operations
 */
void
MP_Global(void *vec, int n, int op_type)
{
    int dim, vprocs, source, dest, np, vp, bytes_per, i, *trial;

    double *dvec = (double *) vec;
    int *ivec = (int *) vec;
    float *fvec = (float *) vec;
    void *work;
    double *dwork;
    int *iwork;
    float *fwork;

    if (nprocs == 1)
	return;

    dmp5("%d in MP_Global(%p, %d, %d)\n", myproc, vec, n, op_type);

    /*  Add pipelining and sync at a later time to overlap comp with comm
    */

    if (op_type % 3 == 1)
        bytes_per = sizeof(int);
    else if (op_type % 3 == 2)
        bytes_per = sizeof(float);
    else
        bytes_per = sizeof(double);

    if (!(work = malloc(n*bytes_per))) {
        fprintf(mylog,"ERROR - malloc(%d) failed in MP_Global()\n",
	  n * bytes_per);
        fflush(mylog);
        exit(1);
    }
    dwork = (double *) work;
    iwork = (int *) work;
    fwork = (float *) work;

    dim = 0; vprocs = 1;
    while (vprocs < nprocs) {
	++dim;
	vprocs *= 2;
    }
    trial = malloc(vprocs * sizeof(int));
    for (i=0; i<nprocs; i++)
	trial[i] = 1;
    for (; i<vprocs; i++)
	trial[i] = 0;

    for (np=1; np<vprocs; np*=2) {
        dest = myproc^np;
        if (dest < nprocs)
	    MP_Send(vec, n*bytes_per, dest);

        if (myproc == nprocs-1)
	    for (vp=nprocs; vp<vprocs; vp++) {
		dest = vp^np;
		if (trial[vp] != 0 && dest < nprocs-1)
		    MP_Send(vec, n*bytes_per, dest);
	    }

        source = myproc^np;
	if (trial[source] != 0) {
	    if (source >= nprocs)
		source = nprocs-1;
	    MP_Recv(work, n*bytes_per, source);
	} else {
	    if (op_type%2 == 0)
		for (i=0; i<n; i++)
		    dwork[i] = 0.0;
	    if (op_type%2 == 1)
		for (i=0; i<n; i++)
		    iwork[i] = 0;
	}
	for (i=0; i<vprocs; i++)
	    trial[i] += trial[i^np];

	dmp4("%d MP_Global() dvec = %p, dwork = %p\n", myproc, dvec, dwork);

	switch(op_type) {
	  case 0:                                        /* MP_dSum() */
	    for (i=0; i<n; i++)
		dvec[i] += dwork[i];
	    break;
	  case 1:                                        /* MP_iSum() */
	    for (i=0; i<n; i++)
		ivec[i] += iwork[i];
	    break;
	  case 2:                                        /* MP_fSum() */
	    for (i=0; i<n; i++)
		fvec[i] += fwork[i];
	    break;
	  case 3:                                        /* MP_dMax() */
	    for (i=0; i<n; i++)
		if (dwork[i] > dvec[i])
		    dvec[i] = dwork[i];
	    break;
	  case 4:                                        /* MP_iMax() */
	    for (i=0; i<n; i++)
		if (iwork[i] > ivec[i])
		    ivec[i] = iwork[i];
	    break;
	  case 5:                                        /* MP_fMax() */
	    for (i=0; i<n; i++)
		if (fwork[i] > fvec[i])
		    fvec[i] = fwork[i];
	    break;
	  case 6:                                        /* MP_dMin() */
	    for (i=0; i<n; i++)
		if (dwork[i] < dvec[i])
		    dvec[i] = dwork[i];
	    break;
	  case 7:                                        /* MP_iMin() */
	    for (i=0; i<n; i++)
		if (iwork[i] < ivec[i])
		    ivec[i] = iwork[i];
	    break;
	  case 8:                                        /* MP_fMin() */
	    for (i=0; i<n; i++)
		if (fwork[i] < fvec[i])
		    fvec[i] = fwork[i];
	    break;
	  default:
	    printf("ERROR - MP_Global() op_type of %d is unkown\n",op_type);
	}
    }
    free(work);
    dmp2("%d exiting MP_Global()\n",myproc);
}

#if 0
void MP_dSum(double *ptr, int nelements)
{ MP_Global(ptr, nelements, 0); }

void MP_iSum(int *ptr, int nelements)
{ MP_Global(ptr, nelements, 1); }

void MP_fSum(float *ptr, int nelements)
{ MP_Global(ptr, nelements, 2); }

void MP_dMax(double *ptr, int nelements)
{ MP_Global(ptr, nelements, 3); }

void MP_iMax(int *ptr, int nelements)
{ MP_Global(ptr, nelements, 4); }

void MP_fMax(float *ptr, int nelements)
{ MP_Global(ptr, nelements, 5); }

void MP_dMin(double *ptr, int nelements)
{ MP_Global(ptr, nelements, 6); }

void MP_iMin(int *ptr, int nelements)
{ MP_Global(ptr, nelements, 7); }

void MP_fMin(float *ptr, int nelements)
{ MP_Global(ptr, nelements, 8); }
#endif


/* Synchronous send and receive - iRecv preposts then sends ready signal
 *   - This guarantees a bypass of any receive buffer
 */

/* definitely broken, msg not int in call to ARecv  --pw */
#if 0
void MP_sSend(void *buf, int nbytes, int dest)
{
   int token;

   MP_Recv(&token, sizeof(int), dest);
   MP_Send(buf, nbytes, dest);
}

void MP_sRecv(void *buf, int nbytes, int src)
{
   int token = 888;
   struct MP_msg_entry msg;

   MP_ARecv(buf, nbytes, src, &msg);
   MP_Send(&token, sizeof(int), src);
   MP_Wait(&msg);
}

void MP_sShift(void *buf_out, int dest, void *buf_in, int src, int nbytes)
{
   int token = 888, nsize, nb, nbmax = (buffer_size/32)*8;
   char *ptr_in, *ptr_out;

   nbmax = 10000;

   ptr_in = buf_in;
   ptr_out = buf_out;
   MP_Send(&token, sizeof(int), src);
   MP_Recv(&token, sizeof(int), dest);
   for( nb=0; nb<nbytes; nb+=nbmax) {
      nsize = MIN(nbytes-nb, nbmax);
/*      if( nbytes > nbmax ) {*/
/*         MP_Send(&token, sizeof(int), src);*/
/*         MP_Recv(&token, sizeof(int), dest);*/
/*      }*/
      MP_Send(ptr_out, nsize, dest);
      MP_Recv(ptr_in, nsize, src);
      ptr_out += nsize;
      ptr_in += nsize;
   }
}


void MP_s2Shift(void *buf_out, int dest, void *buf_in, int src, int nbytes)
{
   int token = 888, npackets, p, packet_size;
   struct MP_msg_entry msg, msg_token;
   char *ptr;

   npackets = (int) ( 0.5 + (4.0*nbytes)/buffer_size );
   if( npackets < 1 || npackets > 1000 ) npackets = 1;

   MP_ARecv(&token, sizeof(int), dest, &msg_token);  /* works for src=dest */

   MP_ARecv(buf_in, nbytes, src, &msg);
   MP_Send(&token, sizeof(int), src);

   MP_Wait(&msg_token);                 /* wait for the ready signal */
/*
   ptr = buf_out;
   packet_size = nbytes / npackets;
   for( p=0; p<npackets; p++) {
      if( p == npackets-1 ) packet_size = nbytes - (npackets-1)*packet_size;
      MP_Send(ptr, packet_size, dest);
      ptr += packet_size;
   }
*/
   MP_Send(buf_out, nbytes, dest);

   MP_Wait(&msg);
}
#endif  /* definitely broken section  --pw */


void
errorw(int err_num, char* text)
{
    fprintf(mylog, "\nWARNING(%d) - %s\n\n", err_num, text); fflush(mylog);
}

void
errorx(int err_num, char* text)
{
    fprintf(stderr,"\nERROR(%d) - %s\n", err_num, text);
    fprintf(mylog, "\nERROR(%d) - %s\n", err_num, text);
    fflush(mylog);
    exit(9);
}

void
errorxs(int err_num, char *format_statement, char *string_ptr)
{
    char format_line[1024];

    strcpy(format_line, "\nERROR(%d) - ");
    strcat(format_line, format_statement);
    strcat(format_line, "\n");
    fprintf(stderr, format_line, err_num, string_ptr);
    fprintf(mylog,  format_line, err_num, string_ptr);
    fflush(mylog);
    exit(9);
}

void
node_range_check(int node)
{
    if (node < 0 || node >= nprocs) {
	const char *err = "ERROR - Send/Recv to/from node %d is out of the"
	  " range 0-%d\n";
	fprintf(stderr, err, node, nprocs-1);
	fprintf(mylog,  err, node, nprocs-1);
	fflush(mylog);
        exit(9);
    }
    if (node == myproc) {
        const char *err = "ERROR - Can't send/recv messages to your own node\n";
        fprintf(stderr, err);
        fprintf(mylog,  err);
        exit(9);
    }
}

#ifdef TRACE
int t_min = 000, t_cur = 0;
double t_0;
struct trace_element t_el[10000];

void
Trace_Dump(void)
{
   int i, n;
   FILE *fd = NULL;
   char filename[100], *arrow;
   double dt;

   strcpy(filename,".trace.");  strcat(filename,myhostname);
   fd=fopen(filename,"w");
   if( !fd ) errorxs(0,"Couldn't open the trace file %s\n",filename);

   n = MIN(t_cur, 10000);
   fprintf(fd,"  Trace Dump from %s of the first %d comms"
              " larger than %d bytes\n\n", myhostname, n, t_min);

   for( i=1; i<n; i++) {
      t_el[i].start -= t_0;
      t_el[i].end -= t_0;

      if( t_el[i].nbytes < 0 ) {             /* receive */
         arrow = "<--";
         t_el[i].nbytes = - t_el[i].nbytes;
      } else if( t_el[i].node >= nprocs ) {  /* buffered send */
         t_el[i].node -= nprocs;
         arrow = "*->";
      } else {                               /* unbuffered send */
         arrow = "-->";
      }

      dt = ( t_el[i].end - t_el[i].start ) * 1000000;

      if( t_el[i].nbytes == 0 ) {
        fprintf(fd,"%9.6f - %9.6f  nd %d ***************************",
           t_el[i].start, t_el[i].end, myproc);
      } else if (t_el[i].nbytes > 0 ) {
        fprintf(fd,"%9.6f - %9.6f  nd %d  %s  nd %d   %8d bytes",t_el[i].start,
            t_el[i].end, myproc, arrow, t_el[i].node, t_el[i].nbytes);
      }
      if( t_el[i].npackets > 1 )
        fprintf(fd," / %d usec (%d pkts)\n", (int) dt, t_el[i].npackets);
      else fprintf(fd," / %d usec\n", (int) dt);
   }
   fclose(fd);
}
#endif  /* TRACE */

