/* mplite.h - header file for using mplite stuff */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#if defined (sgi)
#include <sys/resource.h>
#endif
#include <sys/types.h>
#include <math.h>


#ifdef MPLITE_MAIN     /* Global variables set in mplite.c */

  int myproc ,nprocs;
  FILE *mylog = NULL;  /* .mplite.mymachine file to log the current state */
  char *hostname[1024], myhostname[128], logfilename[128];
  int buffer_size = 1048576;

#else

  extern int myproc ,nprocs;
  extern FILE *mylog;
  extern char *hostname[1024], myhostname[128], logfilename[128];
  extern int buffer_size;

#endif

extern void errorw(int err_num, char *text);
extern void errorx(int err_num, char *text);
extern void errorxs(int err_num, char *format_statement, char *string_ptr);
extern void node_range_check(int node);
#ifdef TRACE
extern void Trace_Dump(void);
#endif

#define MIN(a,b) ((a) < (b) ? (a) : (b))

#define TIMER_RESOLUTION 0.001

#define PANIC_TIME  5.0     /* seconds to wait before logging a warning */
#define USER_FRIENDLY

#define mp1(a) {fprintf(mylog,a);fflush(mylog);}
#define mp2(a,b) {fprintf(mylog,a,b);fflush(mylog);}
#define mp3(a,b,c) {fprintf(mylog,a,b,c);fflush(mylog);}
#define mp4(a,b,c,d) {fprintf(mylog,a,b,c,d);fflush(mylog);}
#define mp5(a,b,c,d,e) {fprintf(mylog,a,b,c,d,e);fflush(mylog);}

#define DEBUG_LEVEL 0

#if DEBUG_LEVEL == 1
  #define dmp1(a) {fprintf(mylog,a);fflush(mylog);}
  #define dmp2(a,b) {fprintf(mylog,a,b);fflush(mylog);}
  #define dmp3(a,b,c) {fprintf(mylog,a,b,c);fflush(mylog);}
  #define dmp4(a,b,c,d) {fprintf(mylog,a,b,c,d);fflush(mylog);}
  #define dmp5(a,b,c,d,e) {fprintf(mylog,a,b,c,d,e);fflush(mylog);}

#elif DEBUG_LEVEL == 2
  #define dmp1(a) {fprintf(mylog,a);printf(a);fflush(mylog);}
  #define dmp2(a,b) {fprintf(mylog,a,b);printf(a,b);fflush(mylog);}
  #define dmp3(a,b,c) {fprintf(mylog,a,b,c);printf(a,b,c);fflush(mylog);}
  #define dmp4(a,b,c,d) {fprintf(mylog,a,b,c,d);printf(a,b,c,d);\
                         fflush(mylog);}
  #define dmp5(a,b,c,d,e) {fprintf(mylog,a,b,c,d,e);printf(a,b,c,d,e);\
                           fflush(mylog);}
#else
  #define dmp1(a) {}
  #define dmp2(a,b) {}
  #define dmp3(a,b,c) {}
  #define dmp4(a,b,c,d) {}
  #define dmp5(a,b,c,d,e) {}
#endif

struct MP_msg_entry {
    void *next;        /* pointer to the next header in a linked list */
    char *buf;         /* Beginning of the buffer */
    char *cur_ptr;     /* Next data to send/recv */
    int nbytes;        /* Total number of bytes to send or receive */
    int nleft;         /* Number of bytes left to send/recv */
    int src;           /* source node number (0..nprocs-1) */
    int dest;          /* destination node number (0..nprocs-1) */
    int send_flag;     /* 1 ==> send, 2 ==> buffer me, 3 ==> buffered */
    int trace_id;
    int stats;
};


struct trace_element {
    int node;
    int nbytes;
    int npackets;
    double start;
    double end;
};

