/*                  pong.c Generic Benchmark code
 *               Dave Turner - Ames Lab - July of 1994+++
 */
#include <stdio.h>
#include <stdlib.h>
#include "mpfuncs.h"

int
main(int argc, char *argv[])
{
    int myproc, end_node, size, other_node, nprocs, i, last;
    double t1, t2, time;
    double *a, *b;
    double max_rate = 0., min_latency = 10.e6;
    int irid, irid_a, irid_b;

    a = (double *) malloc(132000*sizeof(double));
    b = (double *) malloc(132000*sizeof(double));

    for (i=0; i<132000; i++) {
	a[i] = (double) i;
	b[i] = 0.;
    }

    MP_Init(&argc, &argv);
    (void) MP_Myproc(&myproc);
    (void) MP_Nprocs(&nprocs);
    if (nprocs != 2) {
	fprintf(stderr, "pong: run with nprocs == 2\n");
	return 1;
    }
    end_node = nprocs - 1;
    other_node = (myproc + 1) % 2;

    printf("pong: Hello from %d of %d.\n",myproc, nprocs);
    MP_Sync();

    /* Timer accuracy test */
    t1 = MP_Time();
    t2 = MP_Time();
    while (t2 == t1)
	t2 = MP_Time();
    if (myproc == 0)
	printf("Timer accuracy of %f usecs\n", (t2-t1) * 1.0e6);

    /*
     * Communications between nodes 
     *   - Blocking sends and recvs
     *   - No guarantee of prepost, so might pass through comm buffer
     */
    if (myproc==0)
	printf("\n  Blocking ping-pong\n\n");
    for (size=8; size<=524288; size*=2) {
	for (i=0; i<size/8; i++) {
	    a[i] = (double) i;
	    b[i] = 0.0;
	}
        last = size/8 - 1;
	MP_Sync();
	t1 = MP_Time();
	if (myproc == 0) {
	    MP_Send(a, size, other_node);
	    MP_Recv(b, size, other_node);
	} else {
	    MP_Recv(b, size, other_node);
	    b[0] += 1.0;
	    if (last != 0)
		b[last] += 1.0;
	    MP_Send(b, size, other_node);
	}
	t2 = MP_Time();
	time = 1.e6 * (t2 - t1);
	MP_Sync();
	if ((b[0] != 1.0 || b[last] != last + 1)) {
	    printf("ERROR - b[0] = %f b[%d] = %f\n",b[0],last,b[last]);
	    return 1;
	}
	for (i=1; i<last-1; i++)
	if (b[i] != (double) i)
	    printf("ERROR - b[%d] = %f\n", i, b[i]);
	if (myproc == 0 && time > 0.000001) {
	    printf(" %7d bytes took %9.0f usec (%8.3f MB/sec)\n",
	      size,time,2.0*size/time);
	    if (2*size/time > max_rate)
		max_rate = 2*size/time;
	    if (time/2 < min_latency)
		min_latency = time/2;
	} else if (myproc == 0) {
	    printf(" %7d bytes took less than the timer accuracy\n",size);
	}
    }

    /*
     * Async communications
     *   - Prepost receives to guarantee bypassing the comm buffer
     */
    MP_Sync();
    if (myproc==0)
	printf("\n  Asynchronous ping-pong\n\n");
    for  (size=8; size<=524288; size*=2) {
	for (i=0; i<size/8; i++) {
	    a[i] = (double) i;
	    b[i] = 0.0;
	}
	last = size/8 - 1;
	MP_ARecv(b, size, other_node, &irid);
	MP_Sync();
	t1 = MP_Time();
	if (myproc == 0) {
	    MP_Send(a, size, other_node);
	    MP_Wait(&irid);
	} else {
	    MP_Wait(&irid);
	    b[0] += 1.0;
	    if (last != 0)
		b[last] += 1.0;
	    MP_Send(b, size, other_node);
	}
	t2 = MP_Time();
	time = 1.e6 * (t2-t1);
	MP_Sync();
	if ((b[0] != 1.0 || b[last] != last + 1))
	    printf("ERROR - b[0] = %f b[%d] = %f\n",b[0],last,b[last]);
	for (i=1; i<last-1; i++)
	    if (b[i] != (double) i)
		printf("ERROR - b[%d] = %f\n",i, b[i]);
	if (myproc == 0 && time > 0.000001) {
	    printf(" %7d bytes took %9.0f usec (%8.3f MB/sec)\n",
	      size,time,2.0*size/time);
	    if (2*size/time > max_rate)
		max_rate = 2*size/time;
	    if (time/2 < min_latency)
		min_latency = time/2;
	} else if (myproc == 0) {
	    printf(" %7d bytes took less than the timer accuracy\n",size);
	}
    }

    /*
     * Bi-d
     */
    if (myproc==0)
	printf("\n  Bi-directional asynchronous ping-pong\n\n");
    for (size=8; size<=524288; size*=2) {
	for( i=0; i<size/8; i++) {
	    a[i] = (double) i;
	    b[i] = 0.0;
	}
	last = size/8 - 1;
	MP_ARecv(b, size, other_node, &irid_b);
	MP_ARecv(a, size, other_node, &irid_a);
	MP_Sync();
	t1 = MP_Time();

	MP_Send(a, size, other_node);
	MP_Wait(&irid_b);

	b[0] += 1.0;
	if (last != 0)
	    b[last] += 1.0;

	MP_Send(b, size, other_node);
	MP_Wait(&irid_a);

	t2 = MP_Time();
	time = 1.e6 * (t2-t1);
	MP_Sync();
	if ((a[0] != 1.0 || a[last] != last + 1))
	    printf("ERROR - a[0] = %f a[%d] = %f\n",a[0],last,a[last]);
	for (i=1; i<last-1; i++)
	    if (a[i] != (double) i)
		printf("ERROR - a[%d] = %f\n",i, a[i]);
	if (myproc == 0 && time > 0.000001) {
	    printf(" %7d bytes took %9.0f usec (%8.3f MB/sec)\n",
	      size,time,2.0*size/time);
	    if (2*size/time > max_rate)
		max_rate = 2*size/time;
	    if (time/2 < min_latency)
		min_latency = time/2;
	} else if (myproc == 0) {
	    printf(" %7d bytes took less than the timer accuracy\n",size);
	}
    }

    if (myproc == 0)
	printf("\n Max rate = %f MB/sec  Min latency = %f usec\n",
	  max_rate, min_latency);

    MP_Finalize(); 
    return 0;
}

