/* MP_Lite by Dave Turner - Ames Lab - Sept of 1997 +++ */

/* This file is for the TCP version which:
 *   - Uses a .mplite.config file to input nprocs and the hostnames
 *   - Sets up sockets for the basic send and recv commands
 */

#include "mplite.h"   /* Globals myproc, nprocs, mylog, hostname, logfile */

#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#if defined (sgi) || (_AIX)
#endif
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>

#define  TCP_PORT 5820

void MP_Send(), MP_ASend(), MP_Recv(), MP_ARecv(), MP_Wait(), MP_Sync();
void errorx(), errorxs(), errorw();
extern double MP_Time();

#define MAX_MSGS 10000
static int idindex = 0;
struct MP_msg_entry *msg_list[MAX_MSGS];
int n_sends = 0, n_recvs = 0, n_send_packets = 0, n_recv_packets = 0;
int sd[1024], sync_sd_prev, sync_sd_next;   /* socket descriptors */
int salen = sizeof(struct sockaddr_in);
struct sockaddr_in serv_addr;

#if defined (TRACE)
  void Trace_Dump();
  extern int t_min, t_cur;
  extern double t_0;
  extern struct trace_element t_el[10000];
#endif


void MP_Init(int *argc, char ***argv)
{
   FILE *fd;
   int iproc, i, j, n, r, node, next, prev;
   int port = TCP_PORT, sync_port, sync_port_next, listen_socket;
   char *ptr, shortname[100];

   /* Open .mplite.config and read in nprocs and hostnames.
    * Set nprocs and myproc, hostname[0..nprocs-1]
    */

   fd=fopen(".mplite.config","r");
   if( !fd ) errorx(0,"Couldn't open the .mplite.config file\n");

   fscanf(fd,"%d",&nprocs);

   gethostname(myhostname,100);
   ptr = strstr(myhostname,".");
   if (ptr != NULL) strcpy(ptr,"\0");
   strcpy(logfilename,".");  strcat(logfilename,myhostname);
   mylog=fopen(logfilename,"w");
   if( !mylog ) errorxs(0,"Couldn't open the log file %s\n",logfilename);

   for(i=0; i<nprocs; i++) {
      hostname[i] = malloc(100*sizeof(char));
      fscanf(fd,"%d %s\n",&iproc, hostname[i]);
      strcpy(shortname, hostname[i]);
      ptr = strstr(shortname,".");
      if (ptr != NULL) strcpy(ptr,"\0");
      if( !strcmp(shortname,myhostname) ) myproc = i;
/*      if( strstr(hostname[i],myhostname) != NULL ) myproc = i;*/
   }
   if( myproc < 0 ) 
       errorxs(myproc,"%s had no entry in .mplite.config\n",myhostname);

   mp4("%s is node %d of %d\n",myhostname,myproc,nprocs);
   for(i=0; i<nprocs; i++) mp3("   %d ==> %s\n",i,hostname[i]);

   /* Set up a socket for each host.
    * Handshake to each, with a timeout.
    *    Report any errors to the master.
    *    If master is the problem, dump to .mplite.myhostname.
    */

   listen_socket = create_listen_socket(&port);
   sync_port = port+1;

   for(i=0; i<nprocs; i++) {

      if( i < myproc ) {    /* listen for the connection */
         sd[i] = accept_a_connection_from(i, listen_socket);
      }

      /* Wait for go signal from lower iprocs before connecting
       */

      if( i == myproc && myproc > 0 ) {
         n = recv(sd[myproc-1], &j, sizeof(j),0);
         if( n < 0 ) errorx(n,"tcp.c#MP_Init() - send()");
      }

      if( i > myproc ) {             /* initiate the connection */
         sd[i] = connect_to(i, TCP_PORT);
      }
   }

   /* Send 'go' signal for the next iproc to connect */

   if( myproc < nprocs-1 ) {
      n = send(sd[myproc+1], &j, sizeof(j),0);
      if( n < 0 ) errorx(n,"tcp.c#MP_Init() - send() go signal");
   }
   close(listen_socket);


   /* Set up blocking sockets to the next and previous nodes for MP_Sync() */

   if( nprocs > 1 ) {
      mp1("\nSetting up special sockets for MP_Sync()\n");
      next = (myproc+1)%nprocs;
      prev = (myproc+nprocs-1)%nprocs;

mp3("sync_port = %d   (port = %d)\n",sync_port, port);
      listen_socket = create_listen_socket(&sync_port);
      r = -1; while( r <= 0) r = send(sd[prev], &sync_port,      sizeof(int),0);
      r = -1; while( r <= 0) r = recv(sd[next], &sync_port_next, sizeof(int),0);
      if( myproc == 0 ) {
         sync_sd_next = connect_to(next, sync_port_next);
         if( nprocs == 2 ) {
            sync_sd_prev = sync_sd_next;
         } else {
            sync_sd_prev = accept_a_connection_from(prev, listen_socket);
         }
      } else {
         sync_sd_prev = accept_a_connection_from(prev, listen_socket);
         if( nprocs == 2 ) {
            sync_sd_next = sync_sd_prev;
         } else {
            sync_sd_next = connect_to(next, sync_port_next);
         }
      }
      mp2("%d sockets created for MP_Sync()\n\n", nprocs == 2 ? 1 : 2);
      close(listen_socket);
   }

   MP_Sync();   /* Just to test it out */

#if defined (TRACE)
   t_0 = MP_Time();
#endif

   /* Handshake with all, just to be polite */

   for(i=0; i<nprocs; i++) if( i != myproc ) {
      MP_Send(&myproc,  sizeof(int), i);
   }
   for(i=0; i<nprocs; i++) if( i != myproc ) {
      MP_Recv(&node, sizeof(int), i);
      if( node != i ) 
        mp3("WARNING - socket(%d) connected to node %d in MP_Init()\n",i, node);
   }


   n_sends = n_recvs = 0;

   mp1("Done with all the handshaking, exiting MP_Init()\n");
   mp1("------------------------------------------------\n");
}


void MP_Finalize()
{
   int i;

   MP_Sync();
   mp1("------------------------------------------------\n");
   for(i=0; i<nprocs; i++) if (i != myproc) close(sd[i]);
   mp3("%d sends used %d packets\n",n_sends,n_send_packets);
   mp3("%d receives used %d packets\n",n_recvs,n_recv_packets);
   mp1("************************************************\n");
   fclose(mylog);
#if defined (TRACE)
   Trace_Dump();
#endif
}


/* MP_Send()  - Uses TCP level buffering, blocks until copied to TCP packets
 * MP_ASend() - Same since everything must go through the TCP buffer anyway
 * MP_Recv()  - Uses TCP level buffering, blocks until message copied in
 * MP_ARecv() - Sets up msg recv entry, delays anything until MP_Wait()
 * MP_Wait()  - MP_Recv() using *MP_ARecv() entry, or nothing for *MP_ASend()
 */


void MP_Send(void *buf, int nbytes, int dest)
{
   int n, nleft, npackets=0;
   char *ptr;
   ptr = (char *)buf;

dmp4("  Node %d MP_Send(,%d,%d)\n",myproc,nbytes,dest);
   buffersize_warning(nbytes);
   node_range_check(dest);

#if defined (TRACE)
   if( nbytes > t_min && ++t_cur < 10000 ) {
      t_el[t_cur].node = dest;
      t_el[t_cur].nbytes = nbytes;
      t_el[t_cur].start = MP_Time();
   }
#endif

   nleft = nbytes;
   while (nleft > 0) {
      n = write(sd[dest], ptr, nleft);
      if( n < 0 ) errorx(n,"MP_Send()");
      if( n != nbytes ) errorw(n,"MP_Send()");
      npackets++;
      nleft -= n;
      ptr += n;
   }

#if defined (TRACE)
   if( t_cur < 10000 ) {
     t_el[t_cur].end = MP_Time();
     t_el[t_cur].npackets = npackets;
   }
#endif
   n_sends++; n_send_packets += npackets;
}    

void MP_ASend(void *buf, int nbytes, int dest, int *msg_id)
{
   MP_Send(buf, nbytes, dest);
   *msg_id = -1;
   dmp4("%d MP_ASend() done %d %d\n", myproc, nbytes, dest);
}


void MP_Recv(void *buf, int nbytes, int src)
{
   int n, nleft, npackets=0;
   char *ptr;
   ptr = (char *)buf;

dmp4("  Node %d MP_Recv(,%d,%d)\n",myproc,nbytes,src);
   buffersize_warning(nbytes);
   node_range_check(src);

#if defined (TRACE)
   if( nbytes > t_min && ++t_cur < 10000 ) {
      t_el[t_cur].node = src;
      t_el[t_cur].nbytes = -nbytes;
      t_el[t_cur].start = MP_Time();
   }
#endif
   nleft = nbytes;
   while (nleft > 0) {
      n = read(sd[src], ptr, nleft);
      if( n < 0 ) errorx(n,"MP_Recv()");
      if( n == 0 && npackets == 0)
         mp3("MP_Recv(,%d,%d) - 0 bytes read???\n",nbytes,src);
      npackets++;
      nleft -= n;
      ptr += n;
   }

   /* Eventually, set time-out and loop-back around, dumping wait time
    * to the .mplite.host file for easy post-mordem
    */
#if defined (TRACE)
   if( t_cur < 10000 ) {
     t_el[t_cur].end = MP_Time();
     t_el[t_cur].npackets = npackets;
   }
#endif
   n_recvs++; n_recv_packets += npackets;
}

void MP_ARecv(void *buf, int nbytes, int src, int *msg_id)
{
   struct MP_msg_entry *msg;

   buffersize_warning(nbytes);
   node_range_check(src);

   while( msg_list[++idindex%MAX_MSGS] != NULL ) {};
   *msg_id = idindex = idindex%MAX_MSGS;
   msg = msg_list[idindex] = malloc(sizeof(struct MP_msg_entry));
   if( ! msg ) errorx(0,"tcp_sync.c#MP_ASend() - malloc() failed");

   msg->dest_ptr = buf;
   msg->nbytes = nbytes;
   msg->src = src;

   dmp4("%d posted recv %d %d\n", myproc, msg->nbytes, msg->src);
}

void MP_Wait(int *msg_id)
{
   struct MP_msg_entry *msg;

   /* *msg_id = -1 means it was an MP_ASend(), do nothing */

   if( *msg_id >= 0 ) {
      msg = msg_list[*msg_id];
      dmp4("%d MP_Wait() %d %d\n", myproc, msg->nbytes, msg->src);
      MP_Recv(msg->dest_ptr, msg->nbytes, msg->src);
      free(msg);
      msg_list[*msg_id] = NULL;
   }
   dmp1("MP_Wait() done\n");
}

int MP_Nprocs(int* n)
{ 
   *n = nprocs; 
   return nprocs;
}

int MP_Myproc(int* m)
{
   *m = myproc;
   return myproc;
}


/* Barrier sync using separate blocking sockets
 *   - this avoid conflicts with the existing iRecvs so you can for example
 *     post an iRecv and do an MP_Sync() before sending the data to insure
 *     that the iRecv is preposted and the data will bypass the send buffer.
 *   - This is not the fastest sync(), but it is a simple and sure method.
 */

void MP_Sync()
{
   int r1, r2, r3, r4, token = 777, szoi = sizeof(int);

   if( nprocs == 1) return;

   dmp1("MP_Sync() started\n");

#if defined (TRACE)
   if( ++t_cur < 10000 ) {
     t_el[t_cur].nbytes = 0;
     t_el[t_cur].start = MP_Time();
   }
#endif

   if( myproc == 0 )
      r1 = send(sync_sd_next, &token, szoi, 0);

   r2 = recv(sync_sd_prev, &token, szoi, 0);
   r3 = send(sync_sd_next, &token, szoi, 0);
   r4 = recv(sync_sd_prev, &token, szoi, 0);

   if( myproc != 0 )
      r1 = send(sync_sd_next, &token, szoi, 0);

   if( r1 <= 0 || r2 <= 0 || r3 <= 0 || r4 <= 0 )
      errorx(0,"problem in MP_Sync()");

#if defined (TRACE)
   if( t_cur < 10000 ) t_el[t_cur].end = MP_Time();
#endif

   dmp1("MP_Sync() done\n");
}


/* Utility routines for above */

int create_socket()
{
   int sockdr, one = 1, buffer_orig = 0, d = sizeof(int), r;
   static int print_once = 1;

   sockdr = socket(AF_INET, SOCK_STREAM, 0);
   if( sockdr < 0 ) errorx(errno,"MP_Init()-create_socket()");

   /* Set TCP_NODELAY */

   r = setsockopt(sockdr, IPPROTO_TCP, TCP_NODELAY, &one, sizeof(one));
   if( r < 0) errorx(errno,"MP_Init()-setsockopt(NODELAY)");

   /* Attempt to set the send and receive buffer sizes to 4 MB */

   r = getsockopt(sockdr, SOL_SOCKET, SO_RCVBUF,
                  (char *) &buffer_orig,(void *) &d);
   if( r < 0) errorw(errno,"MP_Init()-getsockopt(SO_RCVBUF)");

   r = -1;
   while ( r < 0 && buffer_size > buffer_orig ) {
      r = setsockopt(sockdr, SOL_SOCKET, SO_SNDBUF,
                     (char *) &buffer_size, sizeof(buffer_size));
      r = setsockopt(sockdr, SOL_SOCKET, SO_RCVBUF,
                     (char *) &buffer_size, sizeof(buffer_size));
      if( r < 0 ) buffer_size /= 2;
   }

   getsockopt(sockdr, SOL_SOCKET, SO_RCVBUF, (char *) &buffer_size,(void *) &d);
   if( print_once ) {
      mp3("\nSend/Receive Buffers: %d ==> %d bytes\n\n",
          buffer_orig,buffer_size);
      print_once = 0;
   }

   return sockdr;
}

int create_listen_socket(int* port)
{
   int sockdr, bound = -2, tries = 0;

   sockdr = create_socket();
   serv_addr.sin_family      = AF_INET;
   serv_addr.sin_port        = htons(*port);
   serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
   while( bound < 0 && ++tries < 10) {
      dmp1("Trying to bind for listen_socket\n");
      bound = bind( sockdr, (struct sockaddr *) &serv_addr, salen);
      if( bound < 0 ) {
         mp3("Can't bind to listen port %d - errno = %d\n", *port,errno);
         serv_addr.sin_port   = htons(++ *port);
      }
   }
   if( bound < 0 ) errorx(errno,"MP_Init()-bind()");

   mp2("Bind to listen port %d done - ", *port);

   listen(sockdr, 5);
   mp1("listen completed\n");

   return sockdr;
}


int accept_a_connection_from(int node, int listen_socket)
{
   int sockdr;

   mp2("Trying to accept a connection from %d - ", node);
   bzero( (char *) &serv_addr, sizeof(serv_addr) );
   sockdr = accept(listen_socket, (struct sockaddr *) &serv_addr,
                  (void *) &salen );
   if( sockdr < 0 ) errorx(errno,"MP_Init()-accept()");
   mp1("accepted\n");

   return sockdr;
}


int connect_to(int node, int base_port)
{
   struct hostent *addr;
   int sockdr, port, connected = -2, tries = 0;

   mp2("Trying to connect to node %d - ", node);
   addr = gethostbyname(hostname[node]);
   if( addr == NULL ) errorxs(0,"Invalid hostname %s\n",hostname[node]);

   /* Try to connect for upto 60 seconds before giving up */

   while( connected < 0 && tries++ < 300) {
      serv_addr.sin_family = addr->h_addrtype;
      bcopy(addr->h_addr, &(serv_addr.sin_addr.s_addr), addr->h_length);
      port = base_port + (tries-1)%10;
      serv_addr.sin_port = htons(port);

      /* Alpha Linux needs the socket reset after each try
       * for some reason */

      sockdr = create_socket();
      connected = connect(sockdr, (struct sockaddr *) &serv_addr, salen);

      if (connected < 0) {
        close(sockdr);
        if (errno != ECONNREFUSED) errorx(errno,"MP_Init()-connect()");
        if (tries%10 == 0) {
          printf("Sleeping 2 secs after trying ports %d-%d for node %d\n",
                  TCP_PORT,TCP_PORT+9,node);
          sleep(2);
        }
      }
   }
   if( connected < 0 ) errorx(errno,"MP_Init()-connect()");
   mp3("connected after %d tries on port %d\n", tries,port);

   return sockdr;
}


int buffersize_warning(int nbytes)
{
   static int warn_once = 1;

   if( warn_once && nbytes > buffer_size/2 ) {
      mp3("WARNING - %d byte message being sent through a"
              " %d size buffer\n", nbytes, buffer_size);
      warn_once = 0;
   }
}


void Dump_and_Quit()
{
   mp1("\n\nReceived a SIGINT (cntl-c) --> dump and exit\n");
/*   Dump_Message_State();*/
   fclose(mylog);
   exit(-1);
}

