/*
 * test/vitest.c
 * Copyright (C) 1998 Chris Csanady and Pete Wyckoff
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include "vivar.h"
#include "extern.h"

const char *dev = "/dev/vi0";

void call_trap() {
    int i;
    struct vi_trap_args args;

    args.mh_idx = 6;
    args.vi_idx=9;

/*    for (i = 0; i < 1000000; i++)*/
	VI_TRAP(&args);
}

char *random_user_data = "abcdefghijklmonpqr1234567890\n\0";

char daisy27[6] = { 0x00, 0xa0, 0xc9, 0xc5, 0x83, 0x5b };
char daisy28[6] = { 0x00, 0xa0, 0xc9, 0xc5, 0x81, 0xa7 };

main(int argc, char *argv[]) {
    struct vi_attribs attribs;
    struct vi_mem_attr mem_attribs;
    int nic_handle, vi_handle, mem_handle, pmh, dmh;
    struct vi_test_req req;
    struct vi_cseg  *descs, *desc;
    struct vi_dseg  *dseg;
    char  *packet;
    struct vi_modify_req mreq;
    int i;


    if (VipOpenNic(dev, &nic_handle) < 0)
	printf("ViOpenNic: failed\n");
    if (VipCreateVi(nic_handle, &attribs, 0, 0, &vi_handle) < 0)
	printf("VipCreateVi: failed\n");

    mreq.vi = vi_handle;
    bcopy(daisy27, mreq.ether_dhost, 6);
    mreq.dvi = 7;
    if (ioctl(nic_handle, VI_MODIFY, &mreq) < 0)
	printf("modify failed\n");

    packet = (char *)malloc(2048);
    descs = (struct vi_cseg *)malloc(1024);
    if (packet == NULL || descs == NULL) {
	perror("malloc");
	return;
    }
    bzero(packet, 2048);
    bzero(descs, 1024);

    if (VipRegisterMem(nic_handle, packet, 2048, &mem_attribs, &pmh) < 0)
	perror("VipRegisterMem: packet");
    if (VipRegisterMem(nic_handle, descs, 1024, &mem_attribs, &dmh) < 0)
	perror("VipRegisterMem: descs");

    bcopy(random_user_data, packet, strlen(random_user_data));
    if ((int)packet & 0x3F)
	printf("packet unaligned\n");
    if ((int)descs & 0x3F)
	printf("descs unaligned\n");
    printf("packet mem=%8x, desc mem=%8x\n", packet, descs);

    desc = descs;
    for (i = 0; i < 5; i++) {
	desc->SegCount = 1;
	desc->Control = VIP_CONTROL_OP_SENDRECV;
	desc->Length = strlen(random_user_data);

	dseg = (struct vi_dseg *)(desc + 1);
	dseg->Handle = pmh;
	dseg->Data = packet;
	dseg->Length = strlen(random_user_data);

	if (VipPostSend(vi_handle, desc, dmh) < 0)
	    perror("VipPostSend");

	desc += 2;
    }


    desc = descs;
    for (i = 0; i < 5; i++) {
	while (desc->Status == 0);
	printf("Desc %d done\n", i);
	desc = desc->Next.Address;
    }


    if (VipDeregisterMem(nic_handle, packet, pmh) < 0)
	perror("VipDeregisterMem: packet");
    if (VipDeregisterMem(nic_handle, descs, dmh) < 0)
	perror("VipDeregisterMem: descs");

    if (VipDestroyVi(nic_handle, vi_handle) < 0)
	printf("VipDestroyVi: failed\n");

    if (VipCloseNic(nic_handle) < 0)
	printf("ViCloseNic: failed\n");
}

